""" Display_Widget module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
import numpy as np
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QCheckBox, QComboBox, QHBoxLayout, QVBoxLayout, QMessageBox, QSizePolicy
from magneticalc.QMessageBox2 import QMessageBox2
from magneticalc.Debug import Debug
from magneticalc.QtWidgets2.QGroupBox2 import QGroupBox2
from magneticalc.QtWidgets2.QHLine import QHLine
from magneticalc.QtWidgets2.QIconLabel import QIconLabel
from magneticalc.QtWidgets2.QLabel2 import QLabel2
from magneticalc.QtWidgets2.QSliderFloat import QSliderFloat
from magneticalc.SamplingVolume_Widget import SamplingVolume_Widget
from magneticalc.Theme import Theme
from magneticalc.VisPyCanvas import VisPyCanvas


class Display_Widget(QGroupBox2):
    """ Display_Widget class. """

    # Slider limits
    FieldArrowHeadScaleMin = 0
    FieldArrowHeadScaleMax = 1
    FieldArrowHeadScaleStep = 1 / VisPyCanvas.FieldArrowHeadSize
    FieldArrowLineScaleMin = 0
    FieldArrowLineScaleMax = 1
    FieldArrowLineScaleStep = 1 / 10
    FieldPointScaleMin = 0
    FieldPointScaleMax = 1
    FieldPointScaleStep = 1 / VisPyCanvas.FieldPointSize
    FieldBoostMin = 0
    FieldBoostMax = 1
    FieldBoostStep = 1 / 20

    # Warn about displaying an excessive number of field labels
    ExcessiveFieldLabelsThreshold = 250

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Populates the widget.

        @param gui: GUI
        """
        QGroupBox2.__init__(self, "Display", color=Theme.DarkColor)
        Debug(self, ": Init", init=True)
        self.gui = gui

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.addLayout(QIconLabel("Point Scale", "fa.circle", color=Theme.DarkColor))
        self.field_point_scale_slider = QSliderFloat(
            orientation=Qt.Horizontal,
            minimum=self.FieldPointScaleMin,
            maximum=self.FieldPointScaleMax,
            step=self.FieldPointScaleStep
        )
        self.field_point_scale_slider.valueChanged.connect(  # type: ignore
            lambda: self.set_field_point_scale(self.field_point_scale_slider.get_value())
        )
        self.addWidget(self.field_point_scale_slider)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.addWidget(QHLine())

        self.addLayout(QIconLabel("Arrow Scale", "fa.arrow-right", color=Theme.DarkColor))

        field_arrow_scale_layout_left = QVBoxLayout()
        field_arrow_scale_layout_right = QVBoxLayout()

        field_arrow_scale_layout_left.addWidget(QLabel2("Head:", expand=False))
        self.field_arrow_head_scale_slider = QSliderFloat(
            orientation=Qt.Horizontal,
            minimum=self.FieldArrowHeadScaleMin,
            maximum=self.FieldArrowHeadScaleMax,
            step=self.FieldArrowHeadScaleStep
        )
        self.field_arrow_head_scale_slider.valueChanged.connect(  # type: ignore
            lambda: self.set_field_arrow_head_scale(self.field_arrow_head_scale_slider.get_value())
        )
        field_arrow_scale_layout_right.addWidget(self.field_arrow_head_scale_slider)

        field_arrow_scale_layout_left.addWidget(QLabel2("Line:", expand=False))
        self.field_arrow_line_scale_slider = QSliderFloat(
            orientation=Qt.Horizontal,
            minimum=self.FieldArrowLineScaleMin,
            maximum=self.FieldArrowLineScaleMax,
            step=self.FieldArrowLineScaleStep
        )
        self.field_arrow_line_scale_slider.valueChanged.connect(  # type: ignore
            lambda: self.set_field_arrow_line_scale(self.field_arrow_line_scale_slider.get_value())
        )
        field_arrow_scale_layout_right.addWidget(self.field_arrow_line_scale_slider)

        field_arrow_scale_layout = QHBoxLayout()
        field_arrow_scale_layout.addLayout(field_arrow_scale_layout_left)
        field_arrow_scale_layout.addLayout(field_arrow_scale_layout_right)
        self.addLayout(field_arrow_scale_layout)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.addWidget(QHLine())

        self.addLayout(QIconLabel("Field Boost", "fa.adjust", color=Theme.DarkColor))
        self.field_boost_slider = QSliderFloat(
            orientation=Qt.Horizontal,
            minimum=self.FieldBoostMin,
            maximum=self.FieldBoostMax,
            step=self.FieldBoostStep
        )
        self.field_boost_slider.valueChanged.connect(  # type: ignore
            lambda: self.set_field_boost(self.field_boost_slider.get_value())
        )
        self.addWidget(self.field_boost_slider)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.addWidget(QHLine())

        self.addLayout(QIconLabel("Field Labels", "fa.tags", color=Theme.DarkColor))
        self.display_field_magnitude_labels_checkbox = QCheckBox(" Display Magnitude")
        self.display_field_magnitude_labels_checkbox.toggled.connect(  # type: ignore
            lambda: self.set_display_field_magnitude_labels(self.display_field_magnitude_labels_checkbox.isChecked())
        )
        self.addWidget(self.display_field_magnitude_labels_checkbox)

        self.field_label_resolution_combobox = QComboBox()
        self.field_label_resolution_combobox.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.field_label_resolution_combobox_connection = None
        field_label_resolution_layout = QHBoxLayout()
        field_label_resolution_layout.addWidget(self.field_label_resolution_combobox)
        field_label_resolution_layout.addWidget(QLabel2(" Labels / cm", expand=False))
        self.addLayout(field_label_resolution_layout)

        total_labels_layout = QHBoxLayout()
        self.total_labels_label = QLabel2("N/A", color=Theme.MainColor, align_right=True)
        total_labels_layout.addWidget(QLabel2("Total labels:", italic=True, color=Theme.LiteColor))
        total_labels_layout.addWidget(self.total_labels_label)
        self.addLayout(total_labels_layout)

    def reload(self) -> None:
        """
        Reloads the widget.
        """
        Debug(self, ".reload()", refresh=True)

        self.blockSignals(True)

        self.field_point_scale_slider.setValue(self.gui.project.get_float("field_point_scale"))
        self.field_arrow_head_scale_slider.setValue(self.gui.project.get_float("field_arrow_head_scale"))
        self.field_arrow_line_scale_slider.setValue(self.gui.project.get_float("field_arrow_line_scale"))
        self.field_boost_slider.setValue(self.gui.project.get_float("field_boost"))
        self.display_field_magnitude_labels_checkbox.setChecked(
            self.gui.project.get_bool("display_field_magnitude_labels")
        )

        self.blockSignals(False)

        self.update()

    def update(self) -> None:
        """
        Updates this widget.
        """
        Debug(self, ".update()", refresh=True)

        self.update_labels()
        self.update_controls()

    def update_labels(self) -> None:
        """
        Updates the labels.
        """
        Debug(self, ".update_labels()", refresh=True)

        if self.gui.model.sampling_volume.valid:
            n = self.gui.model.sampling_volume.labels_count
            color = Theme.FailureColor if n > self.ExcessiveFieldLabelsThreshold else Theme.MainColor
            self.total_labels_label.setText(str(n))
            self.total_labels_label.setStyleSheet(f"color: {color}; font-style: italic;")
        else:
            self.total_labels_label.setText("N/A")
            self.total_labels_label.setStyleSheet(f"color: {Theme.LiteColor}; font-style: italic;")

    def update_controls(self) -> None:
        """
        Updates the field label resolution combobox.
        """
        Debug(self, ".update_controls()", refresh=True)

        # Possible field label resolution values: Less than or equal to sampling volume resolution
        sampling_volume_resolution = self.gui.model.sampling_volume.resolution
        label_resolution_options_dict = {
            key: value for key, value in SamplingVolume_Widget.ResolutionOptionsDict.items()
            if np.power(2.0, value) <= sampling_volume_resolution
        }

        self.blockSignals(True)

        # Repopulate field label resolution combobox
        if self.field_label_resolution_combobox_connection is not None:
            self.field_label_resolution_combobox.currentIndexChanged.disconnect(  # type: ignore
                self.field_label_resolution_combobox_connection
            )
        self.field_label_resolution_combobox.clear()
        for i, value in enumerate(label_resolution_options_dict.keys()):
            self.field_label_resolution_combobox.addItem(str(value))

        def connection():
            """ Sets field label resolution. """
            self.set_field_label_resolution(
                label_resolution_options_dict.get(
                    self.field_label_resolution_combobox.currentText(),
                    0
                )
            )
        self.field_label_resolution_combobox_connection = connection
        self.field_label_resolution_combobox.currentIndexChanged.connect(connection)  # type: ignore

        # Set default field label resolution if it is not available anymore
        target = self.gui.project.get_int("sampling_volume_label_resolution_exponent")
        if target not in label_resolution_options_dict.values():
            Debug(self, f": WARNING: Invalid: sampling_volume_label_resolution_exponent = {target}", warning=True)
            self.gui.project.set_int(
                "sampling_volume_label_resolution_exponent",
                next(iter(label_resolution_options_dict.items()))[1]  # First value from combobox
            )

        # Select the field label resolution
        target = self.gui.project.get_int("sampling_volume_label_resolution_exponent")
        for i, value in enumerate(label_resolution_options_dict.values()):
            if value == target:
                self.field_label_resolution_combobox.setCurrentIndex(i)

        self.blockSignals(False)

    # ------------------------------------------------------------------------------------------------------------------

    def set_field_point_scale(self, value: float) -> None:
        """
        Sets field point scale.

        @param value: Value
        """
        if self.signalsBlocked():
            return

        self.gui.project.set_float("field_point_scale", value)
        self.gui.redraw()

    def set_field_arrow_head_scale(self, value: float) -> None:
        """
        Sets field arrow head scale.

        @param value: Value
        """
        if self.signalsBlocked():
            return

        self.gui.project.set_float("field_arrow_head_scale", value)
        self.gui.redraw()

    def set_field_arrow_line_scale(self, value: float) -> None:
        """
        Sets field arrow line scale.

        @param value: Value
        """
        if self.signalsBlocked():
            return

        self.gui.project.set_float("field_arrow_line_scale", value)
        self.gui.redraw()

    def set_field_boost(self, value: float) -> None:
        """
        Sets field boost value.

        @param value: Value
        """
        if self.signalsBlocked():
            return

        self.gui.project.set_float("field_boost", value)
        self.gui.redraw()

    # ------------------------------------------------------------------------------------------------------------------

    def set_display_field_magnitude_labels(self, value: bool) -> None:
        """
        Sets field label "Display Magnitude" value.

        @param value: Value
        """
        if self.signalsBlocked():
            return

        self.gui.project.set_bool("display_field_magnitude_labels", value)

        self.prevent_excessive_field_labels(choice=True)

        self.update()

        self.gui.redraw()

    def set_field_label_resolution(self, value: int) -> None:
        """
        Sets field label resolution exponent.

        @param value: Value
        """
        if self.signalsBlocked():
            return

        self.gui.sidebar_left.sampling_volume_widget.set_sampling_volume(_label_resolution_exponent_=value)

        # Note: "prevent_excessive_field_labels()" will be called by "Model.on_sampling_volume_valid()".

        self.gui.redraw()

    def disable_field_labels(self) -> None:
        """
        Disables field labels.
        """
        Debug(self, ".disable_field_labels()")

        self.gui.project.set_bool("display_field_magnitude_labels", False)

        previous_signals_blocked = self.signalsBlocked()
        self.blockSignals(True)
        self.display_field_magnitude_labels_checkbox.setChecked(False)
        self.blockSignals(previous_signals_blocked)

    def prevent_excessive_field_labels(self, choice: bool) -> None:
        """
        Prevents displaying an excessive number of field labels.

        @param choice: True lets the user choose; False disables field labels if there is an excessive number of them
        """
        Debug(self, f".prevent_excessive_field_labels(choice={choice})")

        if not self.gui.project.get_bool("display_field_magnitude_labels"):
            return

        if not self.gui.model.sampling_volume.valid:
            return

        if not self.gui.model.sampling_volume.labels_count > self.ExcessiveFieldLabelsThreshold:
            return

        if choice:
            text = (
                "You are about to display an excessive number of field labels.\n"
                "This will be very slow and cannot be interrupted.\n\n"

                "DO YOU WANT TO DISPLAY FIELD LABELS ANYWAY?\n"
                "Choosing 'No' will disable field labels immediately.\n\n"

                "Please consider the following before choosing 'Yes':\n"
                "– Save your work first.\n"
                "– Decrease sampling volume resolution.\n"
                "– Decrease field label resolution."
            )
            messagebox = QMessageBox2(
                title="Excessive Number Of Field Labels",
                text=text,
                icon=QMessageBox.Question,
                buttons=QMessageBox.Yes | QMessageBox.No,
                default_button=QMessageBox.No
            )
            if not messagebox.user_accepted or messagebox.choice == QMessageBox.No:
                self.disable_field_labels()

        else:

            self.disable_field_labels()

            text = (
                "Field labels were disabled automatically because\n"
                "an excessive number of field labels was detected.\n\n"

                "You can manually re-enable field labels\n"
                "after all calculations have finished."
            )
            QMessageBox2(
                title="Excessive Number Of Field Labels",
                text=text,
                icon=QMessageBox.Information,
                buttons=QMessageBox.Ok,
                default_button=QMessageBox.Ok
            )



""" Validatable module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Any, Callable
from magneticalc.Assert_Dialog import Assert_Dialog
from magneticalc.Debug import Debug


class Validatable(object):
    """ Validatable class. """

    def __init__(self) -> None:
        """
        Initializes a validatable object.
        """
        Debug(self, ": Init", init=True)
        self._valid = False

    @property
    def valid(self) -> bool:
        """
        @return: True if valid, False if invalid
        """
        return self._valid

    @valid.setter
    def valid(self, valid: bool) -> None:
        """
        @param valid: True if valid, False if invalid
        """
        Debug(self, f".valid = {valid}", success=valid, warning=not valid)
        self._valid = valid


def require_valid(func: Callable) -> Callable:
    """
    Decorator that requires a validatable object to be valid before calling a function.

    @param func: Method of a validatable instance
    @return: Function
    """

    def wrapper(validatable: Validatable, *args, **kwargs) -> Any:
        """
        Wrapper that requires a validatable object to be valid.

        @param validatable: Validatable instance
        @return: Result of the function
        """
        if not validatable.valid:
            Assert_Dialog(False, "Attempting to access invalidated object")
        result = func(validatable, *args, **kwargs)
        return result

    return wrapper


def validator(func: Callable) -> Callable:
    """
    Decorator that invalidates a validatable object and then validates it based on the result of calling a function.

    @param func: Method of a validatable instance
    @return: Function
    """

    def wrapper(validatable: Validatable, *args, **kwargs) -> Any:
        """
        Wrapper that invalidates a validatable object and then validates it based on the result of calling a function.

        @param validatable: Validatable instance
        @return: Result of the function
        """
        validatable.valid = False
        result = func(validatable, *args, **kwargs)
        validatable.valid = result
        return result

    return wrapper



""" QDoubleSpinBox2 module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Callable
from PyQt5.QtWidgets import QDoubleSpinBox, QSizePolicy


class QDoubleSpinBox2(QDoubleSpinBox):
    """ QDoubleSpinBox2 class. """

    def __init__(
            self,
            gui: GUI,  # type: ignore
            minimum: float,
            maximum: float,
            step: float,
            precision: int,
            value: int,
            value_changed: Callable
    ) -> None:
        """
        Initializes a floating-point spinbox.

        @param gui: GUI
        @param minimum: Minimum value
        @param maximum: Maximum value
        @param step: Step
        @param precision: Precision
        @param value: Initial value
        @param value_changed: Value changed callback
        """
        QDoubleSpinBox.__init__(self)
        self.setLocale(gui.user_locale)
        self.setMinimum(minimum)
        self.setMaximum(maximum)
        self.setValue(value)
        self.setSingleStep(step)
        self.setDecimals(precision)
        self.valueChanged.connect(value_changed)  # type: ignore
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)



""" Config module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Dict, List, Optional, Union
import os
import abc
import configparser
import numpy as np
from magneticalc.Debug import Debug
from magneticalc.Format import Format


class Config(metaclass=abc.ABCMeta):
    """ Config class. """

    """ Enables debug output for configuration access. """
    DebugAccess = False

    def __init__(self) -> None:
        """
        Initializes the configuration.
        """
        Debug(self, ": Init", init=True)
        self._parser: Optional[configparser.ConfigParser] = None
        self.filename: Optional[str] = None
        self.synced: bool = False

    @abc.abstractmethod
    def on_changed(self) -> None:
        """
        Gets called when the configuration was loaded, saved or changed.
        """
        pass

    def load(self, default_config: Dict, signal_changed: bool = True) -> None:
        """
        Loads the default configuration.

        @param default_config: Default configuration
        """
        Debug(self, f".load(): Loading defaults")
        self._parser = configparser.ConfigParser()
        self._parser["DEFAULT"] = default_config

        Debug(self, ".load(): Creating empty User section")
        self._parser["User"] = {}

        self.synced = True

        if signal_changed:
            self.on_changed()

    def load_file(self, filename: str, default_config: Dict) -> None:
        """
        Loads the configuration from file.

        @param filename: Filename
        @param default_config: Default configuration
        """
        self.filename = Format.absolute_filename(filename)
        Debug(self, f".load_file(): {Format.filename_uri(self.filename)}")

        self.load(default_config, signal_changed=False)

        assert self._parser is not None, "Not initialized"

        self._parser.read(self.filename)

        if "User" not in self._parser:
            Debug(self, ".load_file(): Creating empty User section")
            self._parser["User"] = {}
            self.synced = False

        if not os.path.isfile(self.filename):
            Debug(
                self, f".load_file(): WARNING: File does not exist: {Format.filename_uri(self.filename)}", warning=True
            )
            self.synced = False

        self.on_changed()

    def save_file(self, filename: Optional[str] = None) -> None:
        """
        Saves the configuration to file.

        @param filename: Filename (optional)
        """
        assert self._parser is not None, "Not initialized"

        if filename is not None:
            self.filename = Format.absolute_filename(filename)

        assert self.filename is not None, "No filename"

        Debug(self, f".save_file(): {Format.filename_uri(self.filename)}")

        with open(self.filename, "w") as file:
            self._parser.write(file)

        self.synced = True
        self.on_changed()

    def cleanup(self) -> None:
        """
        Perform clean-up.
        """
        assert self._parser is not None, "Not initialized"

        Debug(self, ".cleanup()")
        self._parser = None
        self.filename = None
        self.synced = False

    # ------------------------------------------------------------------------------------------------------------------

    def remove_key(self, key: str) -> None:
        """
        Removes a key from the configuration.

        @param key: Key
        """
        assert self._parser is not None, "Not initialized"

        if not self._parser.remove_option("User", key):
            Debug(self, f".remove_key({key}): ERROR: No such key", error=True)

    # ------------------------------------------------------------------------------------------------------------------

    def set_get_str(self, key: str, value: str) -> str:
        """
        If value is not None, writes and returns key-value; otherwise (if value is None), reads and returns key-value.

        @param key: Key
        @param value: Value
        """
        if value is None:
            return self.get_str(key)
        else:
            self.set_str(key, value)
            return value

    def set_str(self, key: str, value: str) -> None:
        """
        Writes a configuration value.

        @param key: Key
        @param value: Value
        """
        assert self._parser is not None, "Not initialized"

        old_value = self._parser.get("User", key, fallback=None)
        if old_value != value:
            if self.DebugAccess:
                if old_value is None:
                    Debug(self, f".set_str(\"{key}\"): Created key with value \"{value}\"")
                else:
                    Debug(self, f".set_str(\"{key}\"): Changing \"{old_value}\" to \"{value}\"")
            self._parser.set("User", key, value)
            self.synced = False
            self.on_changed()
        else:
            if self.DebugAccess:
                Debug(self, f".set_str(\"{key}\"): Unchanged: \"{old_value}\"")

    def get_str(self, key: str) -> str:
        """
        Reads a configuration value.

        @param key: Key
        @return: Value
        """
        assert self._parser is not None, "Not initialized"

        value = self._parser.get("User", key, fallback=None)
        assert value is not None, f"Attempting to read non-existing key: {key}"

        if self.DebugAccess:
            Debug(self, f".get_str(\"{key}\") = \"{value}\"")

        return value

    # ------------------------------------------------------------------------------------------------------------------

    def set_get_bool(self, key: str, value: bool) -> bool:
        """
        If value is not None, writes and returns key-value; otherwise (if value is None), reads and returns key-value.

        @param key: Key
        @param value: Value
        """
        if value is None:
            return self.get_bool(key)
        else:
            self.set_bool(key, value)
            return value

    def set_bool(self, key: str, value: bool) -> None:
        """
        Sets boolean value, convert to string.

        @param key: Key
        @param value: Value
        """
        self.set_str(key, "True" if value else "False")

    def get_bool(self, key: str) -> bool:
        """
        Gets boolean value, convert from string.

        @param key: Key
        @return: Value
        """
        return self.get_str(key) == "True"

    # ------------------------------------------------------------------------------------------------------------------

    def set_get_int(self, key: str, value: int) -> int:
        """
        If value is not None, writes and returns key-value; otherwise (if value is None), reads and returns key-value.

        @param key: Key
        @param value: Value
        """
        if value is None:
            return self.get_int(key)
        else:
            self.set_int(key, value)
            return value

    def set_int(self, key: str, value: int) -> None:
        """
        Sets integer value, convert to string.

        @param key: Key
        @param value: Value
        """
        self.set_str(key, str(int(value)))

    def get_int(self, key: str) -> int:
        """
        Gets integer value, convert from string.

        @param key: Key
        @return: Value
        """
        return int(self.get_str(key))

    # ------------------------------------------------------------------------------------------------------------------

    def set_get_float(self, key: str, value: float) -> float:
        """
        If value is not None, writes and returns key-value; otherwise (if value is None), reads and returns key-value.

        @param key: Key
        @param value: Value
        """
        if value is None:
            return self.get_float(key)
        else:
            self.set_float(key, value)
            return value

    def set_float(self, key: str, value: float) -> None:
        """
        Sets float value, convert to string.

        @param key: Key
        @param value: Value
        """
        self.set_str(key, Format.float_to_str(value))

    def get_float(self, key: str) -> float:
        """
        Gets float value, convert from string.

        @param key: Key
        @return: Value
        """
        return float(self.get_str(key))

    # ------------------------------------------------------------------------------------------------------------------

    def set_get_point(
            self,
            key: str,
            value: Union[np.ndarray, List[float]]
    ) -> Union[np.ndarray, List[float]]:
        """
        If value is not None, writes and returns key-value; otherwise (if value is None), reads and returns key-value.

        @param key: Key
        @param value: Value
        """
        if value is None:
            return self.get_point(key)
        else:
            self.set_point(key, value)
            return value

    def set_point(self, key: str, value: Union[np.ndarray, List[float]]) -> None:
        """
        Sets a single 3D point, convert to string.

        @param key: Key
        @param value: Single 3D point
        """
        self.set_str(key, Format.point_to_str(value))

    def get_point(self, key: str) -> List[float]:
        """
        Gets a single 3D point, convert from string.

        @param key: Key
        @return: Single 3D point
        """
        return Format.str_to_point(self.get_str(key))

    # ------------------------------------------------------------------------------------------------------------------

    def set_get_points(
            self,
            key: str,
            value: Union[np.ndarray, List[List[float]]]
    ) -> Union[np.ndarray, List[List[float]]]:
        """
        If value is not None, writes and returns key-value; otherwise (if value is None), reads and returns key-value.

        @param key: Key
        @param value: Value
        """
        if value is None:
            return self.get_points(key)
        else:
            self.set_points(key, value)
            return value

    def set_points(self, key: str, value: Union[np.ndarray, List[List[float]]]) -> None:
        """
        Sets list of 3D points, convert to string.

        @param key: Key
        @param value: List of 3D points
        """
        self.set_str(key, Format.points_to_str(value))

    def get_points(self, key: str) -> List[List[float]]:
        """
        Gets list of 3D points, convert from string.

        @param key: Key
        @return: List of 3D points
        """
        return Format.str_to_points(self.get_str(key))

    # ------------------------------------------------------------------------------------------------------------------

    def set_get_dict(self, prefix: str, suffix: str, types: Dict, values: Optional[Dict]) -> Dict:
        """
        If "values" is None, reads and returns all key-values in "types".
        Note: The actual keys saved in the configuration file are prefixed with "prefix", and suffixed with "suffix".

        If "values" is not None, writes and returns all key-values in "types".
        Note: In this case, "values" must have the same keys as "types".

        @param prefix: Prefix
        @param suffix: Suffix
        @param types: Key:Type (Dictionary)
        @param values: Key:Value (Dictionary)
        """
        if values is None:
            result = {}
            for _key, _type in types.items():
                result[_key] = self.get_generic(prefix + _key + suffix, _type)
            return result
        else:
            for _key, _type in types.items():
                self.set_generic(prefix + _key + suffix, _type, values[_key])
            return values

    def get_generic(
            self,
            _key: str,
            _type: str
    ) -> Union[str, bool, int, float, np.ndarray, List[float], List[List[float]]]:
        """
        Gets some "_key"-"_value" of data type "_type".

        @param _key: Key
        @param _type: Data type
        @return: Data (type determined by "_type")
        """
        set_func = {
            "bool"  : self.get_bool,
            "float" : self.get_float,
            "int"   : self.get_int,
            "point" : self.get_point,
            "points": self.get_points,
            "str"   : self.get_str
        }

        if _type in set_func:
            return set_func[_type](_key)
        else:
            raise TypeError("Invalid type")

    def set_generic(
            self,
            _key: str,
            _type: str,
            _value: Union[str, bool, int, float, np.ndarray, List[float], List[List[float]]]
    ) -> None:
        """
        Sets some "_key"-"_value" of data type "_type".

        @param _key: Key
        @param _type: Data type
        @param _value: Data (type determined by "_type")
        """
        set_func = {
            "bool"  : self.set_bool,
            "float" : self.set_float,
            "int"   : self.set_int,
            "point" : self.set_point,
            "points": self.set_points,
            "str"   : self.set_str
        }

        if _type in set_func:
            set_func[_type](_key, _value)
        else:
            raise TypeError("Invalid type")



""" Norm_Types module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from magneticalc.Assert_Dialog import Assert_Dialog


""" Norm type: X. """
NORM_TYPE_X = 0

""" Norm type: Y. """
NORM_TYPE_Y = 1

""" Norm type: Z. """
NORM_TYPE_Z = 2

""" Norm type: Radius. """
NORM_TYPE_RADIUS = 3

""" Norm type: Radius X. """
NORM_TYPE_RADIUS_X = 4

""" Norm type: Radius Y. """
NORM_TYPE_RADIUS_Y = 5

""" Norm type: Radius Z. """
NORM_TYPE_RADIUS_Z = 6

""" Norm type: Radius XY. """
NORM_TYPE_RADIUS_XY = 7

""" Norm type: Radius XZ. """
NORM_TYPE_RADIUS_XZ = 8

""" Norm type: Radius YZ. """
NORM_TYPE_RADIUS_YZ = 9

""" Norm type: Angle XY. """
NORM_TYPE_ANGLE_XY = 10

""" Norm type: Angle XZ. """
NORM_TYPE_ANGLE_XZ = 11

""" Norm type: Angle YZ. """
NORM_TYPE_ANGLE_YZ = 12

""" Norm type: Divergence. """
NORM_TYPE_DIVERGENCE = 13


""" Map of norm types to names. """
Norm_Types_Names_Map = {
    NORM_TYPE_X             : "X",
    NORM_TYPE_Y             : "Y",
    NORM_TYPE_Z             : "Z",
    NORM_TYPE_RADIUS        : "Radius",
    NORM_TYPE_RADIUS_X      : "Radius X",
    NORM_TYPE_RADIUS_Y      : "Radius Y",
    NORM_TYPE_RADIUS_Z      : "Radius Z",
    NORM_TYPE_RADIUS_XY     : "Radius XY",
    NORM_TYPE_RADIUS_XZ     : "Radius XZ",
    NORM_TYPE_RADIUS_YZ     : "Radius YZ",
    NORM_TYPE_ANGLE_XY      : "Angle XY",
    NORM_TYPE_ANGLE_XZ      : "Angle XZ",
    NORM_TYPE_ANGLE_YZ      : "Angle YZ",
    NORM_TYPE_DIVERGENCE    : "Divergence",
}

""" Default norm type. """
Norm_Type_Default = NORM_TYPE_X


def norm_type_safe(norm_type: int) -> int:
    """
    A valid norm type is passed through, but an invalid norm type converts to the default type.

    @param norm_type: Norm type
    @return: Safe norm type
    """
    if norm_type in Norm_Types_Names_Map:
        return norm_type
    else:
        Assert_Dialog(False, "Invalid norm type, using fallback type")
        return Norm_Type_Default


def norm_type_to_name(norm_type: int) -> str:
    """
    Converts a norm type to a norm name.

    @param norm_type: Norm type
    @return: Norm name
    """
    return Norm_Types_Names_Map[norm_type_safe(norm_type)]


def norm_name_to_type(norm_name: str) -> int:
    """
    Converts a norm name to a norm type.

    @param norm_name: Norm name
    @return Norm type
    """
    return {_name: _type for _type, _name in Norm_Types_Names_Map.items()}[norm_name]



""" Format module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import List, Union
import os
import numpy as np


class Format:
    """ Format class. """

    """ Float precision. """
    FloatPrecision = 4

    """ Coordinate precision. """
    CoordinatePrecision = 6

    @staticmethod
    def absolute_filename(filename: str) -> str:
        """
        Converts a filename to an absolute filename.

        @param filename: Filename
        @return: Absolute filename
        """
        return os.path.join(os.path.dirname(os.path.dirname(__file__)), filename)

    @staticmethod
    def filename_uri(filename: str) -> str:
        """
        Converts a filename to a filename URI.

        @param filename: Filename
        @return: Filename URI
        """
        return "file://" + filename.replace(" ", "%20")

    @staticmethod
    def float_to_str(value: float) -> str:
        """
        Converts a float value to string.
        """
        return f"{float(value):.{Format.FloatPrecision}f}"

    @staticmethod
    def point_to_str(point: Union[np.ndarray, List[float]]) -> str:
        """
        Converts a single 3D point to string.

        @param point: Single 3D point
        @return: String
        """
        return ", ".join([f"{component:+0.0{Format.CoordinatePrecision}f}" for component in point])

    @staticmethod
    def points_to_str(points: Union[np.ndarray, List[List[float]]]) -> str:
        """
        Converts list of 3D points to string.

        @param points: List of 3D points
        @return: String
        """
        return "; ".join([Format.point_to_str(point) for point in points])

    @staticmethod
    def str_to_point(str_point: str) -> Union[np.ndarray, List[float]]:
        """
        Converts string to single 3D point.

        @param str_point: String
        @return: Single 3D point
        """
        return [float(coord) for coord in str_point.split(",")]

    @staticmethod
    def str_to_points(str_points: str) -> Union[np.ndarray, List[List[float]]]:
        """
        Converts string to list of 3D points.

        @param str_points: String
        @return: List of 3D points
        """
        return [
            Format.str_to_point(str_point) for str_point in str_points.split(";")
            if str_point != ""  # Ignoring stray semicolon at the end of the line
        ]



""" Field module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Tuple, Callable
import numpy as np
from numba import jit, prange, set_num_threads
from magneticalc.Assert_Dialog import Assert_Dialog
from magneticalc.Backend_Types import BACKEND_TYPE_JIT, BACKEND_TYPE_CUDA, get_jit_enabled
from magneticalc.Backend_CUDA import Backend_CUDA
from magneticalc.Backend_JIT import Backend_JIT
from magneticalc.ConditionalDecorator import ConditionalDecorator
from magneticalc.Debug import Debug
from magneticalc.Field_Types import FIELD_TYPE_A, FIELD_TYPE_B
from magneticalc.SamplingVolume import SamplingVolume
from magneticalc.Validatable import Validatable, require_valid, validator
from magneticalc.Wire import Wire


class Field(Validatable):
    """ Field class. """

    def __init__(self) -> None:
        """
        Initializes a field.
        """
        Validatable.__init__(self)
        Debug(self, ": Init", init=True)

        # Parameters
        self.type = 0
        self._distance_limit = 0
        self._length_scale = 0

        # Results
        self._total_calculations: int = 0
        self._total_skipped_calculations: int = 0
        self._vectors: np.ndarray = np.array([])

    def set(
            self,
            type: int,
            distance_limit: float,
            length_scale: float
    ):
        """
        Sets the parameters.

        @param type: Field type
        @param distance_limit: Distance limit (mitigating divisions by zero)
        @param length_scale: Length scale (m)
        """
        self.type = type
        self._distance_limit = distance_limit
        self._length_scale = length_scale

    def get_units(self, show_gauss=False) -> Tuple[str, float]:
        """
        Gets field units.

        @param show_gauss: Enable to show Gauss instead of Tesla
        @return: Field units, field factor
        """
        if show_gauss:
            return {
                FIELD_TYPE_A: "G·m",    # Gauss · meter
                FIELD_TYPE_B: "G"       # Gauss
            }.get(self.type, ""), 1e4
        else:
            return {
                FIELD_TYPE_A: "T·m",    # Tesla · meter
                FIELD_TYPE_B: "T"       # Tesla
            }.get(self.type, ""), 1e0

    # ------------------------------------------------------------------------------------------------------------------

    @validator
    def recalculate(
            self,
            wire: Wire,
            sampling_volume: SamplingVolume,
            progress_callback: Callable,
            num_cores: int,
            backend_type: int
    ) -> bool:
        """
        Recalculates field vectors.

        @param wire: Wire
        @param sampling_volume: Sampling volume
        @param progress_callback: Progress callback
        @param num_cores: Number of cores to use for multiprocessing
        @param backend_type: Backend type
        @return: True if successful, False if interrupted (CUDA backend currently not interruptable)
        """
        Debug(self, ".recalculate()")

        # Compute the current elements.
        current_elements = wire.elements

        if backend_type == BACKEND_TYPE_JIT:

            # Initialize Biot-Savart JIT backend
            backend = Backend_JIT(
                field_type=self.type,
                distance_limit=self._distance_limit,
                length_scale=self._length_scale,
                dc=wire.dc,
                current_elements=current_elements,
                sampling_volume_points=sampling_volume.points,
                sampling_volume_permeabilities=sampling_volume.permeabilities,
                progress_callback=progress_callback
            )

            # Fetch result using Biot-Savart JIT backend
            set_num_threads(num_cores)
            backend_result = backend.get_result()

        elif backend_type == BACKEND_TYPE_CUDA:

            # Initialize Biot-Savart CUDA backend
            backend = Backend_CUDA(
                field_type=self.type,
                distance_limit=self._distance_limit,
                length_scale=self._length_scale,
                dc=wire.dc,
                current_elements=current_elements,
                sampling_volume_points=sampling_volume.points,
                sampling_volume_permeabilities=sampling_volume.permeabilities,
                progress_callback=progress_callback
            )

            # Fetch result using Biot-Savart CUDA backend
            set_num_threads(num_cores)
            backend_result = backend.get_result()

        else:

            Debug(self, f".recalculate(): ERROR: No such backend: {backend_type}", error=True)
            return False

        # Handle interrupt
        if backend_result is None:
            return False

        self._total_calculations = backend_result[0]
        self._total_skipped_calculations = backend_result[1]
        self._vectors = backend_result[2]

        # Sanity check
        expected_total_calculations = len(current_elements) * sampling_volume.points_count
        if expected_total_calculations != self._total_calculations:
            Assert_Dialog(False, "ERROR: Unexpected number of calculations – Backend seems to be buggy")
            return False

        return True

    @property
    @require_valid
    def vectors(self) -> np.ndarray:
        """
        Gets field vectors. (The selected field type determined which field was calculated.)

        @return: Ordered list of 3D vectors (field vectors & corresponding sampling volume points have the same indices)
        """
        return self._vectors

    @property
    @require_valid
    def vectors_count(self) -> int:
        """
        @return: Number of field vectors
        """
        return len(self._vectors)

    @property
    @require_valid
    def total_calculations(self) -> int:
        """
        Gets total number of calculations.

        @return: Total number of calculations
        """
        return self._total_calculations

    @property
    @require_valid
    def total_skipped_calculations(self) -> int:
        """
        Gets total number of skipped calculations.

        @return: Total number of skipped calculations
        """
        return self._total_skipped_calculations

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    @ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=True)
    def get_arrows(
            sampling_volume_points: np.ndarray,
            field_vectors: np.ndarray,
            line_pairs: np.ndarray,
            head_points: np.ndarray,
            arrow_scale: float,
            magnitude_limit: float
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns the field arrow parameters needed by L{VisPyCanvas}.

        @param sampling_volume_points: Sampling volume points
        @param field_vectors: Field vectors
        @param line_pairs: Arrow line pairs (ordered list of arrow start/stop 3D points)
        @param head_points: Arrow head points (ordered list of arrow stop 3D points)
        @param arrow_scale: Arrow scale
        @param magnitude_limit: Magnitude limit (mitigating divisions by zero)
        @return: Line pairs, head points
        """
        for i in prange(len(field_vectors)):

            # Calculate field vector magnitude (mitigating divisions by zero)
            field_vector_length = np.sqrt(
                field_vectors[i][0] ** 2 + field_vectors[i][1] ** 2 + field_vectors[i][2] ** 2
            )
            if field_vector_length < magnitude_limit:
                field_vector_length = magnitude_limit

            # Calculate normalized field direction
            field_direction_norm = field_vectors[i] / field_vector_length

            # Calculate arrow start & end coordinates
            p_start = sampling_volume_points[i] + field_direction_norm / 2 / 2 * arrow_scale
            p_end = sampling_volume_points[i] - field_direction_norm / 2 / 2 * arrow_scale

            # Populate arrow line & head coordinates
            line_pairs[2 * i + 0] = p_start
            line_pairs[2 * i + 1] = p_end
            head_points[i] = p_end

        return line_pairs, head_points



""" About_Dialog module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import webbrowser
from functools import partial
from PyQt5.Qt import QShowEvent
from magneticalc.QtWidgets2.QDialog2 import QDialog2
from magneticalc.QtWidgets2.QTextBrowser2 import QTextBrowser2
from magneticalc.Debug import Debug
from magneticalc.Theme import Theme
from magneticalc.Version import Version


class About_Dialog(QDialog2):
    """ About_Dialog class. """

    # Donation URL
    DonationURL = "https://www.paypal.com/cgi-bin/webscr?cmd=_s-xclick&hosted_button_id=TN6YTPVX36YHA&source=url"

    # HTML content
    HTML = f"""
        <span style="color: {Theme.MainColor};"><b>{Version.String}</b></span><br>
        <br>
        Copyright © 2020–2022, Paul Wilhelm, M. Sc.
        &lt;<a href="mailto:anfrage@paulwilhelm.de">anfrage@paulwilhelm.de</a>&gt;<br>
        <br>
        <small>
            <b>ISC License</b><br>
            <br>
            Permission to use, copy, modify, and/or distribute this software for any
            purpose with or without fee is hereby granted, provided that the above
            copyright notice and this permission notice appear in all copies.<br>
            <br>
            THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
            WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
            MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
            ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
            WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
            ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
            OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.<br>
        </small>
        <br>
        <span style="color: {Theme.MainColor}; font-weight: bold;">
            If you like this software, please consider buying me a coffee!
        </span>
        <br>
        """

    def __init__(self) -> None:
        """
        Initializes the dialog.
        """
        QDialog2.__init__(self, title="About", width=640)
        Debug(self, ": Init", init=True)

        self.text_browser = QTextBrowser2(html=self.HTML)
        self.addWidget(self.text_browser)

        buttons = self.addButtons({
            "OK"            : ("fa.check", self.accept),
            "Donate 3€ …"   : ("fa.paypal", partial(webbrowser.open, About_Dialog.DonationURL))
        })
        buttons[0].setFocus()

    def showEvent(self, event: QShowEvent) -> None:
        """
        Gets called when the dialog is opened.

        @param event: QShowEvent
        """
        Debug(self, ".showEvent()")
        self.text_browser.fit_to_contents()



""" Config_Collection module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Optional, List, Dict
from magneticalc.Assert_Dialog import Assert_Dialog


class Config_Collection:
    """ Config_Collection class. """

    def __init__(
            self,
            gui: GUI,  # type: ignore
            prefix: str,
            types: Dict,
            first_without_suffix: bool
    ) -> None:
        """
        Initializes a config collection: A list of groups (dictionaries).

        Example for the SOMETHING collection, consisting of two groups _0 and _1, each with two keys FOO and BAR:
        SOMETHING_count = 2
        SOMETHING_FOO_0 = "1st value in 1st group"
        SOMETHING_BAR_0 = "2nd value in 1st group"
        SOMETHING_FOO_1 = "1st value in 2nd group"
        SOMETHING_BAR_1 = "2nd value in 2nd group"

        @param gui: GUI
        @param prefix: Prefix
        @param types: Key:Type (Dictionary)
        @param first_without_suffix: Enable to omit the suffix "_0" for the first group
        """
        self.gui = gui
        self.prefix = prefix
        self.types = types
        self.first_without_suffix = first_without_suffix

    def _get_suffix(self, group_index: int) -> str:
        """
        Gets the suffix for a given group index.

        @param group_index: Group index
        @return: String
        """
        return "" if group_index == 0 and self.first_without_suffix else f"_{group_index}"

    def get_count(self) -> int:
        """
        Gets the number of groups in the collection.

        @return: Number of groups in the collection
        """
        return self.gui.project.get_int(self.prefix + "count")

    def get_group(self, group_index: int) -> Dict:
        """
        Gets a group from the collection.

        @param group_index: Group index
        @return: Group (dictionary)
        """
        return self.gui.project.set_get_dict(
            prefix=self.prefix,
            suffix=self._get_suffix(group_index),
            types=self.types,
            values=None
        )

    def get_all_groups(self) -> List[Dict]:
        """
        Gets every group in the collection.

        @return: List of groups (dictionaries)
        """
        return [self.get_group(group_index) for group_index in range(self.get_count())]

    def add_group(self, values: Optional[Dict]) -> None:
        """
        Appends a group to the collection.

        @param values: Key:Value (Dictionary)
        """
        group_index = self.get_count()
        self.gui.project.set_get_dict(
            prefix=self.prefix,
            suffix=self._get_suffix(group_index),
            types=self.types,
            values=values,
        )
        self.gui.project.set_int(self.prefix + "count", group_index + 1)

    def del_group(self, group_index: int) -> None:
        """
        Deletes a group from the collection.

        @param group_index: Group index
        """
        Assert_Dialog(group_index < self.get_count(), "Attempting to delete non-existing group")

        # Make a copy of all groups
        groups = self.get_all_groups()

        # Delete the desired group from the copy
        del groups[group_index]

        # Delete all groups from the collection
        for i in range(self.get_count()):
            for key in self.types.keys():
                self.gui.project.remove_key(f"{self.prefix}{key}_{i}")

        # Clear the group count of the collection
        self.gui.project.set_int(self.prefix + "count", 0)

        # Recreate the groups of the collection (regenerate suffixes)
        for group in groups:
            self.add_group(group)

    def set(self, group_index: int, key: str, value) -> None:
        """
        Sets a value for some key in a group of the collection.

        @param group_index: Group index
        @param key: Key
        @param value: Value
        """
        _type = self.types.get(key)
        self.gui.project.set_generic(f"{self.prefix}{key}_{group_index}", _type, value)



""" Field_Widget module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Optional
from functools import partial
from PyQt5.QtWidgets import QHBoxLayout, QVBoxLayout, QSizePolicy
from PyQt5.QtWidgets import QButtonGroup, QCheckBox, QRadioButton
from magneticalc.QtWidgets2.QDoubleSpinBox2 import QDoubleSpinBox2
from magneticalc.QtWidgets2.QGroupBox2 import QGroupBox2
from magneticalc.QtWidgets2.QHLine import QHLine
from magneticalc.QtWidgets2.QIconLabel import QIconLabel
from magneticalc.QtWidgets2.QLabel2 import QLabel2
from magneticalc.Debug import Debug
from magneticalc.Field_Types import Field_Types_Names_Map
from magneticalc.Metric import Metric
from magneticalc.ModelAccess import ModelAccess
from magneticalc.Theme import Theme


class Field_Widget(QGroupBox2):
    """ Field_Widget class. """

    # Spinbox limits
    DistanceLimitMin = 0.0001
    DistanceLimitMax = 1
    DistanceLimitStep = 0.0001
    DistanceLimitPrecision = 4

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Populates the widget.

        @param gui: GUI
        """
        QGroupBox2.__init__(self, "Field", color=Theme.DarkColor)
        Debug(self, ": Init", init=True)
        self.gui = gui

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.addLayout(QIconLabel("Type", "mdi.tune-variant", color=Theme.DarkColor))

        field_type_layout_left = QVBoxLayout()
        field_type_layout_right = QVBoxLayout()

        self.field_type_group = QButtonGroup()
        field_type_radiobuttons = {}
        self.field_type_checkboxes = {}
        for field_type, field_name in Field_Types_Names_Map.items():
            field_type_radiobuttons[field_type] = QRadioButton(" " + field_name)
            field_type_layout_left.addWidget(field_type_radiobuttons[field_type])
            self.field_type_checkboxes[field_type] = QCheckBox(" Cached")
            self.field_type_checkboxes[field_type].setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
            self.field_type_checkboxes[field_type].setEnabled(False)
            field_type_layout_right.addWidget(self.field_type_checkboxes[field_type])
            self.field_type_group.addButton(field_type_radiobuttons[field_type])

        field_type_layout = QHBoxLayout()
        field_type_layout.addLayout(field_type_layout_left)
        field_type_layout.addLayout(field_type_layout_right)
        self.addLayout(field_type_layout)

        for i, button in enumerate(self.field_type_group.buttons()):
            button.toggled.connect(partial(self.on_field_type_changed, i))

        total_calculations_layout = QHBoxLayout()
        self.total_calculations_label = QLabel2("N/A", color=Theme.MainColor, align_right=True)
        total_calculations_layout.addWidget(QLabel2("Total calculations:", italic=True, color=Theme.LiteColor))
        total_calculations_layout.addWidget(self.total_calculations_label)
        self.addLayout(total_calculations_layout)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.addWidget(QHLine())

        self.addLayout(QIconLabel("Distance Limit", "mdi.ruler", color=Theme.DarkColor))
        self.distance_limit_spinbox = QDoubleSpinBox2(
            gui=self.gui,
            minimum=self.DistanceLimitMin,
            maximum=self.DistanceLimitMax,
            step=self.DistanceLimitStep,
            precision=self.DistanceLimitPrecision,
            value=0,
            value_changed=lambda: self.set_field(_distance_limit_=self.distance_limit_spinbox.value())
        )
        distance_limit_layout = QHBoxLayout()
        distance_limit_layout.addWidget(self.distance_limit_spinbox)
        distance_limit_layout.addWidget(QLabel2("cm", expand=False))
        self.addLayout(distance_limit_layout)

        total_skipped_calculations_layout = QHBoxLayout()
        self.total_skipped_calculations_label = QLabel2("N/A", color=Theme.MainColor, align_right=True)
        total_skipped_calculations_layout.addWidget(
            QLabel2("Total skipped calculations:", italic=True, color=Theme.LiteColor)
        )
        total_skipped_calculations_layout.addWidget(self.total_skipped_calculations_label)
        self.addLayout(total_skipped_calculations_layout)

    def reload(self) -> None:
        """
        Reloads the widget.
        """
        Debug(self, ".reload()", refresh=True)

        self.blockSignals(True)

        field_type = self.gui.project.get_int("field_type")
        for i, button in enumerate(self.field_type_group.buttons()):
            button.setChecked(i == field_type)

        self.distance_limit_spinbox.setValue(self.gui.project.get_float("field_distance_limit"))

        self.blockSignals(False)

        # Initially load field from project
        self.set_field(recalculate=False, invalidate=False)

        self.update()

    def update(self) -> None:
        """
        Updates this widget.
        """
        Debug(self, ".update()", refresh=True)

        self.update_labels()
        self.update_controls()

    def update_labels(self) -> None:
        """
        Updates the labels.
        """
        if self.gui.model.field.valid:
            self.total_calculations_label.setText(str(self.gui.model.field.total_calculations))
            self.total_skipped_calculations_label.setText(str(self.gui.model.field.total_skipped_calculations))
        else:
            self.total_calculations_label.setText("N/A")
            self.total_skipped_calculations_label.setText("N/A")

    def update_controls(self) -> None:
        """
        Updates the controls.
        """
        for field_type in Field_Types_Names_Map:
            self.field_type_checkboxes[field_type].setChecked(self.gui.model.get_valid_field(field_type) is not None)

        self.set_color(Theme.MainColor if self.gui.model.field.valid else Theme.FailureColor)

    # ------------------------------------------------------------------------------------------------------------------

    def on_field_type_changed(self, field_type: int, checked: bool) -> None:
        """
        Gets called when the field type changed.

        @param field_type: Field type
        @param checked: Boolean
        """
        if self.signalsBlocked():
            return

        if not checked:
            return

        self.set_field(_field_type_=field_type, invalidate=False)

    # ------------------------------------------------------------------------------------------------------------------

    def set_field(
            self,
            _field_type_: Optional[int] = None,
            _distance_limit_: Optional[float] = None,
            invalidate: bool = True,
            recalculate: bool = True
    ) -> None:
        """
        Sets the field. This will replace the currently set field in the model.
        Any underscored parameter may be left set to None in order to load its default value.

        @param _field_type_: Field type
        @param _distance_limit_: Distance limit
        @param invalidate: Enable to invalidate this model hierarchy level
        @param recalculate: Enable to trigger final recalculation
        """
        if self.signalsBlocked():
            return

        Debug(self, ".set_field()")

        with ModelAccess(self.gui, recalculate):

            field_type = self.gui.project.set_get_int("field_type", _field_type_)
            distance_limit = self.gui.project.set_get_float("field_distance_limit", _distance_limit_)

            self.gui.model.set_field(
                invalidate=invalidate,
                field_type=field_type,
                distance_limit=distance_limit,
                length_scale=Metric.LengthScale
            )



""" Field_Types module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from magneticalc.Assert_Dialog import Assert_Dialog


""" Field type: A-Field. """
FIELD_TYPE_A = 0

""" Field type: B-Field. """
FIELD_TYPE_B = 1


""" Map of field types to names. """
Field_Types_Names_Map = {
    FIELD_TYPE_A    :   "A-Field (Vector Potential)",
    FIELD_TYPE_B    :   "B-Field (Flux Density)"
}

""" Map of field types to abbreviations. """
Field_Types_Abbreviations_Map = {
    FIELD_TYPE_A    :   "A",
    FIELD_TYPE_B    :   "B"
}

""" Default field type. """
Field_Type_Default = FIELD_TYPE_A


def field_type_safe(field_type: int) -> int:
    """
    A valid field type is passed through, but an invalid field type converts to the default type.

    @param field_type: Field type
    @return: Safe field type
    """
    if field_type in Field_Types_Names_Map:
        return field_type
    else:
        Assert_Dialog(False, "Invalid field type, using fallback type")
        return Field_Type_Default


def field_name_to_type(field_name: str) -> int:
    """
    Converts a field name to a field type.

    @param field_name: Field name
    @return: Field type
    """
    return {_name: _type for _type, _name in Field_Types_Names_Map.items()}[field_name]



""" QButtons module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Dict, Tuple, Callable
import qtawesome as qta
from PyQt5.QtWidgets import QHBoxLayout, QPushButton


class QButtons(QHBoxLayout):
    """ QButtons class. """

    def __init__(
            self,
            data: Dict[str, Tuple[str, Callable]]
    ) -> None:
        """
        Initializes a horizontal layout containing multiple buttons.

        @param data: Dictionary {text: (icon, callback)}
        """
        QHBoxLayout.__init__(self)

        self.dictionary: Dict[int, QPushButton] = {}

        for n, key_val in enumerate(data.items()):
            text, tup = key_val
            icon, callback = tup

            button = QPushButton(qta.icon(icon), " " + text)
            button.clicked.connect(callback)  # type: ignore
            self.addWidget(button)

            self.dictionary[n] = button



""" View_Menu module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from PyQt5.QtWidgets import QMenu
from magneticalc.QCheckboxConfig import QCheckboxConfig
from magneticalc.Debug import Debug


class View_Menu(QMenu):
    """ View_Menu class. """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ):
        """
        Initializes the menu.
        """
        QMenu.__init__(self, "&View")
        Debug(self, ": Init", init=True)
        self.gui = gui

        self.checkbox_config = QCheckboxConfig(gui=self.gui)
        self.addAction(
            self.checkbox_config.add("Show Wire Segments", "show_wire_segments", self.gui.redraw)
        )
        self.addAction(
            self.checkbox_config.add("Show Wire Points", "show_wire_points", self.gui.redraw)
        )
        self.addSeparator()
        self.addAction(
            self.checkbox_config.add("Show Colored Labels", "show_colored_labels", self.gui.redraw)
        )
        self.addAction(
            self.checkbox_config.add("Show Gauss instead of Tesla", "show_gauss", self.on_show_gauss_changed)
        )
        self.addSeparator()
        self.addAction(
            self.checkbox_config.add("Show Coordinate System", "show_coordinate_system", self.gui.redraw)
        )
        self.addAction(
            self.checkbox_config.add("Show Perspective Info", "show_perspective_info", self.gui.redraw)
        )
        self.addSeparator()
        self.addAction(
            self.checkbox_config.add("Dark Background", "dark_background", self.gui.redraw)
        )

    def update(self):
        """
        Updates the menu.
        """
        Debug(self, ".update()", refresh=True)

        self.gui.blockSignals(True)
        self.checkbox_config.reload()
        self.gui.blockSignals(False)

    # ------------------------------------------------------------------------------------------------------------------

    def on_show_gauss_changed(self) -> None:
        """
        Gets called when the "Show Gauss instead of Tesla" option changed.
        """
        self.gui.sidebar_right.metric_widget.update()
        self.gui.vispy_canvas.delete_field_labels()
        self.gui.redraw()



#!/bin/usr/env python3

""" MagnetiCalc main module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
import sys
from sty import ef
from PyQt5.QtWidgets import QApplication
from magneticalc.Backend_Types import get_jit_enabled, get_cuda_available
from magneticalc.GUI import GUI
from magneticalc.Version import Version


def main() -> None:
    """ MagnetiCalc main function. """

    if sys.platform == "win32":
        os.system("color")

    print()
    print(ef.bold + Version.String + ef.rs)
    print(Version.Copyright)
    print(Version.License)
    print()
    print("Performance Options: JIT " + ("Enabled" if get_jit_enabled() else "Disabled") + ", ", end="")
    print("CUDA " + (("Available" if get_cuda_available() else "Not Available") if get_jit_enabled() else "Disabled"))
    print()

    app = QApplication(sys.argv)

    gui = GUI()

    gui.show()

    rc = app.exec()

    sys.exit(rc)


if __name__ == "__main__":
    main()



""" File_Menu module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
import qtawesome as qta
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QMenu, QFileDialog
from magneticalc.QtWidgets2.QSaveAction import QSaveAction
from magneticalc.Debug import Debug
from magneticalc.ExportContainer_Dialog import ExportContainer_Dialog


class File_Menu(QMenu):
    """ File_Menu class. """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ):
        """
        Initializes the menu.
        """
        QMenu.__init__(self, "&File")
        Debug(self, ": Init", init=True)
        self.gui = gui

        self.addAction(
            qta.icon("fa.file"),
            "&New Project …",
            self.new_project,
            Qt.CTRL + Qt.Key_N
        )

        self.addAction(
            qta.icon("fa.folder"),
            "&Open Project …",
            self.open_project,
            Qt.CTRL + Qt.Key_O
        )

        self.addAction(
            qta.icon("fa.save"),
            "&Save Project",
            lambda: self.save_project_as() if self.gui.project.filename is None else self.gui.project.save_file(),
            Qt.CTRL + Qt.Key_S
        )

        self.addAction(
            qta.icon("fa.save"),
            "Save Project &As …",
            self.save_project_as,
            Qt.CTRL + Qt.SHIFT + Qt.Key_S
        )

        self.addAction(
            qta.icon("fa.times"),
            "Close Project",
            self.close_project,
            Qt.CTRL + Qt.Key_C
        )

        self.addSeparator()

        self.addAction(
            qta.icon("fa.picture-o"),
            "Export &Image …",
            self.export_image,
            Qt.CTRL + Qt.Key_I
        )

        self.addAction(
            qta.icon("fa.folder"),
            "&Export Container …",
            lambda: ExportContainer_Dialog(self.gui).show(),
            Qt.CTRL + Qt.Key_E
        )

        self.addSeparator()

        self.addAction(
            qta.icon("fa.window-close"),
            "&Quit",
            self.gui.close,
            Qt.CTRL + Qt.Key_Q
        )

    def update(self):
        """
        Updates the menu.
        """
        Debug(self, ".update()", refresh=True)

    # ------------------------------------------------------------------------------------------------------------------

    def new_project(self) -> None:
        """
        Creates a new project with an INI file.
        """
        Debug(self.gui, ".new_project()")

        if self.gui.calculation_thread is not None:
            self.gui.interrupt_calculation()

        action = QSaveAction(
            self.gui,
            title="New Project",
            date=True,
            filename="MagnetiCalc_Project",
            extension=".ini",
            _filter="MagnetiCalc Project (*.ini)"
        )
        if action.filename:
            self.gui.project.switch(action.filename)
            self.gui.project.save_file()

    def open_project(self) -> None:
        """
        Opens a project from an INI file.
        """
        Debug(self.gui, ".open_project()")

        if self.gui.calculation_thread is not None:
            self.gui.interrupt_calculation()

        filename, _chosen_extension = QFileDialog.getOpenFileName(
            parent=self.gui,
            caption="Open Project",
            filter="MagnetiCalc Project (*.ini)",
            options=QFileDialog.DontUseNativeDialog
        )

        if filename != "":
            self.gui.project.switch(filename)

    def save_project_as(self) -> None:
        """
        Saves the project to an INI file.
        """
        Debug(self.gui, ".save_project_as()")

        action = QSaveAction(
            self.gui,
            title="Save Project",
            date=True,
            filename="MagnetiCalc_Project",
            extension=".ini",
            _filter="MagnetiCalc Project (*.ini)"
        )
        if action.filename:
            self.gui.project.save_file(filename=action.filename)

    def close_project(self) -> None:
        """
        Closes the project and loads the default project.
        """
        Debug(self.gui, ".close_project()")

        self.gui.project.close()
        self.gui.project.open_default()

    def export_image(self) -> None:
        """
        Exports the currently displayed scene to a PNG file.
        """
        Debug(self.gui, ".export_image()")

        action = QSaveAction(
            self.gui,
            title="Save Image",
            date=True,
            filename="MagnetiCalc_Screenshot",
            extension=".png",
            _filter="PNG (*.png)"
        )
        if action.filename:
            self.gui.vispy_canvas.save_image(action.filename)



""" OverridePadding_Dialog module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from magneticalc.QtWidgets2.QDialog2 import QDialog2
from magneticalc.QtWidgets2.QHBoxLayout2 import QHBoxLayout2
from magneticalc.QtWidgets2.QLabel2 import QLabel2
from magneticalc.QtWidgets2.QSpinBox2 import QSpinBox2
from magneticalc.Debug import Debug
from magneticalc.Theme import Theme


class OverridePadding_Dialog(QDialog2):
    """ OverridePadding_Dialog class. """

    # Spinbox limits
    BoundsRange = (-1000, +1000)

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Initializes the dialog.

        @param gui: GUI
        """
        QDialog2.__init__(self, title="Override Padding", width=420)
        Debug(self, ": Init", init=True)
        self.gui = gui

        self.addWidget(
            QLabel2("Please specify the sampling volume bounding box", bold=True, color=Theme.MainColor)
        )
        self.addSpacing(8)
        self.addLayout(QHBoxLayout2(QLabel2("Units:", expand=False), QLabel2("cm", bold=True)))
        self.addSpacing(16)

        bounding_box = self.gui.project.get_points("sampling_volume_bounding_box")
        self.bounds_min_spinbox = [
            QSpinBox2(self.BoundsRange[0], self.BoundsRange[1], int(bounding_box[0][i]), self.validate)
            for i in range(3)
        ]
        self.bounds_max_spinbox = [
            QSpinBox2(self.BoundsRange[0], self.BoundsRange[1], int(bounding_box[1][i]), self.validate)
            for i in range(3)
        ]

        for i in range(3):
            text = "  ≤  " + ["X", "Y", "Z"][i] + "  ≤  "
            self.addLayout(
                QHBoxLayout2(self.bounds_min_spinbox[i], QLabel2(text, expand=False), self.bounds_max_spinbox[i])
            )

        self.addSpacing(16)

        self.buttons = self.addButtons({
            "Cancel": ("fa.close", self.reject),
            "Apply": ("fa.save", self.accept),
        })
        self.buttons[1].setFocus()
        self.validate()

    def validate(self) -> None:
        """
        Validates the bounding box values and enables/disables the "Apply" button accordingly.
        """
        self.buttons[1].setEnabled(
            all([self.bounds_min_spinbox[i].value() <= self.bounds_max_spinbox[i].value() for i in range(3)])
        )



""" Parameters module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Callable
import numpy as np
from numba import jit, prange
from magneticalc.Backend_Types import get_jit_enabled
from magneticalc.ConditionalDecorator import ConditionalDecorator
from magneticalc.Constants import Constants
from magneticalc.Debug import Debug
from magneticalc.Field_Types import FIELD_TYPE_A, FIELD_TYPE_B
from magneticalc.Metric import Metric
from magneticalc.Validatable import Validatable, require_valid, validator


class Parameters(Validatable):
    """ Parameters class. """

    def __init__(self) -> None:
        """
        Initializes parameters class.
        """
        Validatable.__init__(self)
        Debug(self, ": Init", init=True)

        self._energy: float = 0.0
        self._self_inductance: float = 0.0
        self._magnetic_dipole_moment: float = 0.0

    def set(self, *args, **kwargs):
        """
        Sets the parameters
        """

    # ------------------------------------------------------------------------------------------------------------------

    def _get_squared_field(
            self,
            sampling_volume: SamplingVolume,  # type: ignore
            field: Field  # type: ignore
    ) -> float:
        """
        Returns the "squared" field scalar.

        @param sampling_volume: SamplingVolume
        @param field: B-field
        @return: Float
        """
        return self._get_squared_field_worker(sampling_volume.permeabilities, field.vectors)

    @staticmethod
    @ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=True)
    def _get_squared_field_worker(sampling_volume_permeabilities: np.ndarray, field_vectors: np.ndarray) -> float:
        """
        Returns the "squared" field scalar.

        @param sampling_volume_permeabilities: Ordered list of sampling volume's relative permeabilities µ_r
        @param field_vectors: Ordered list of 3D vectors (B-field)
        @return: Float
        """
        squared = 0
        for i in prange(len(field_vectors)):
            squared += np.dot(field_vectors[i], field_vectors[i] / sampling_volume_permeabilities[i])
        return squared

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def _get_magnetic_dipole_moment(
            self, wire: Wire,  # type: ignore
            length_scale: float
    ) -> float:
        """
        Returns the magnetic dipole moment scalar.

        @param wire: Wire
        @param length_scale: Length scale (m)
        @return: Float
        """
        elements_center = np.array([element[0] for element in wire.elements])
        elements_direction = np.array([element[1] for element in wire.elements])
        vector = self._get_magnetic_dipole_moment_worker(elements_center, elements_direction, length_scale)
        return np.abs(wire.dc * np.linalg.norm(vector) / 2)

    @staticmethod
    @ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=True)
    def _get_magnetic_dipole_moment_worker(
            elements_center: np.ndarray,
            elements_direction: np.ndarray,
            length_scale: float
    ):
        """
        Returns the (unscaled) magnetic dipole moment vector.

        @param elements_center: Current element centers
        @param elements_direction: Current element directions
        @param length_scale: Length scale (m)
        @return: Magnetic dipole moment vector
        """
        squared = np.zeros(3)
        for i in prange(len(elements_center)):
            squared += np.cross(elements_center[i] * length_scale, elements_direction[i] * length_scale)
        return squared

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    @validator
    def recalculate(
            self,
            wire: Wire,  # type: ignore
            sampling_volume: SamplingVolume,  # type: ignore
            field: Field,  # type: ignore
            progress_callback: Callable
    ) -> bool:
        """
        Recalculates parameters.

        @param wire: Wire
        @param sampling_volume: SamplingVolume
        @param field: Field
        @param progress_callback: Progress callback
        @return: True (currently non-interruptable)
        """
        Debug(self, ".recalculate()")

        progress_callback(0)

        self._magnetic_dipole_moment = self._get_magnetic_dipole_moment(wire, Metric.LengthScale)

        progress_callback(33)

        if field.type == FIELD_TYPE_A:

            pass

        elif field.type == FIELD_TYPE_B:

            dV = (Metric.LengthScale / sampling_volume.resolution) ** 3  # Sampling volume element
            self._energy = self._get_squared_field(sampling_volume, field) * dV / Constants.mu_0

            progress_callback(66)

            self._self_inductance = self._energy / np.square(wire.dc)

        progress_callback(100)

        return True

    @property
    @require_valid
    def energy(self) -> float:
        """
        Returns calculated energy.

        @return: Float
        """
        return self._energy

    @property
    @require_valid
    def self_inductance(self) -> float:
        """
        Returns calculated self-inductance.

        @return: Float
        """
        return self._self_inductance

    @property
    @require_valid
    def magnetic_dipole_moment(self) -> float:
        """
        Returns calculated magnetic dipole moment.

        @return: Float
        """
        return self._magnetic_dipole_moment



""" Project module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Dict
from PyQt5.QtWidgets import QMessageBox
from magneticalc.QMessageBox2 import QMessageBox2
from magneticalc.Backend_Types import BACKEND_TYPE_CUDA
from magneticalc.Backend_Types import Backend_Types_Available, Backend_Type_Default, backend_type_safe
from magneticalc.Config import Config
from magneticalc.Debug import Debug
from magneticalc.Field_Types import FIELD_TYPE_B, field_type_safe
from magneticalc.Format import Format
from magneticalc.ModelAccess import ModelAccess
from magneticalc.Perspective_Presets import Perspective_Presets
from magneticalc.Version import Version
from magneticalc.Wire_Presets import Wire_Presets


class Project(Config):
    """ Project class. """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ):
        """
        Initializes the project class.

        @param gui: GUI
        """
        Config.__init__(self)
        Debug(self, ": Init", init=True)
        self.gui = gui

    def open_default(self, reload: bool = True) -> None:
        """
        Opens the default project.

        @param reload: Enable to reload GUI
        """
        Debug(self, ".open_default()")
        self.load(default_config=Project.get_default())
        self.validate()

        if reload:
            self.gui.reload()

    def open(self, filename: str, reload: bool = True) -> None:
        """
        Opens a project from file.

        @param filename: Project filename
        @param reload: Enable to reload GUI
        """
        Debug(self, ".open()")
        self.load_file(filename=filename, default_config=Project.get_default())
        self.validate()

        if reload:
            self.gui.reload()

    def close(self, invalidate: bool = True) -> bool:
        """
        Attempts to close the project, but lets user choose to cancel closing, or save/discard changes if there are any.

        @param invalidate: Enable to invalidate the GUI afterwards
        @return: False if canceled, True if saved/discarded
        """
        if self.gui.calculation_thread is not None:
            self.gui.interrupt_calculation()

        if not self.synced:
            Debug(self, ".close(): Project has unsaved changes", warning=True)

            messagebox = QMessageBox2(
                title="Project Changed",
                text="Do you want to save your changes?",
                icon=QMessageBox.Question,
                buttons=QMessageBox.Save | QMessageBox.Discard | QMessageBox.Cancel,
                default_button=QMessageBox.Save
            )
            if not messagebox.user_accepted or messagebox.choice == QMessageBox.Cancel:
                Debug(self, ".close(): Canceled")
                return False
            elif messagebox.choice == QMessageBox.Save:
                Debug(self, ".close(): Saving changes to project", success=True)
                self.save_file()
            else:
                Debug(self, ".close(): Discarding changes to project", warning=True)

        if invalidate:
            Debug(self, ".close(): Invalidating GUI")
            with ModelAccess(self.gui, recalculate=False):
                self.gui.model.invalidate(do_all=True)
        else:
            Debug(self, ".close(): Not invalidating GUI")

        self.cleanup()
        Debug(self, ".close(): Project closed")
        return True

    def switch(self, filename: str) -> None:
        """
        Closes the current project and opens another project.

        @param filename: Project filename
        """
        Debug(self, f".switch(): {filename}")

        if not self.close():
            Debug(self, ".switch(): Canceled")
            return

        self.open(filename)

    # ------------------------------------------------------------------------------------------------------------------

    def on_changed(self) -> None:
        """
        Gets called when the project changed.
        """
        name = self.filename or "Default Project"
        self.gui.setWindowTitle(Version.String + " – " + name + ("" if self.synced else " *"))

    # ------------------------------------------------------------------------------------------------------------------

    def validate(self) -> None:
        """
        Validates the project.
        """
        Debug(self.gui, ".validate()")

        # Ensure valid backend type
        backend_type = self.gui.project.get_int("backend_type")
        if backend_type != backend_type_safe(backend_type):
            backend_type = self.gui.project.set_get_int("backend_type", backend_type_safe(backend_type))

        # Use default backend if selected backend is not available
        if backend_type != Backend_Type_Default and not Backend_Types_Available[backend_type]:
            Debug(self.gui, ".validate(): WARNING: Selected backend not available, using default backend", warning=True)
            self.gui.project.set_int("backend_type", Backend_Type_Default)

        # Ensue valid field type
        field_type = self.gui.project.get_int("field_type")
        if field_type != field_type_safe(field_type):
            self.gui.project.set_int("field_type", field_type_safe(field_type))

    @staticmethod
    def get_default() -> Dict:
        """
        Gets the default project.

        @return: Dictionary
        """
        return {
            "version"                                   : Version.String,
            "backend_type"                              : BACKEND_TYPE_CUDA,
            "auto_calculation"                          : "True",
            "num_cores"                                 : "0",
            "wire_points_base"                          :
                Format.points_to_str(Wire_Presets.get_by_id("Straight Line")["points"]),
            "wire_stretch"                              : "0.1000, 1.0000, 1.0000",
            "wire_slicer_limit"                         : "0.0500",
            "wire_dc"                                   : "1.0000",
            "wire_close_loop"                           : "True",
            "rotational_symmetry_count"                 : "30",
            "rotational_symmetry_radius"                : "1.0000",
            "rotational_symmetry_axis"                  : "2",
            "rotational_symmetry_offset"                : "0.0",
            "sampling_volume_padding"                   : "-1, 1, 1",
            "sampling_volume_override_padding"          : "False",
            "sampling_volume_bounding_box"              : "0.000000, 0.000000, 0.000000; 0.000000, 0.000000, 0.000000",
            "sampling_volume_resolution_exponent"       : "4",
            "sampling_volume_label_resolution_exponent" : "0",
            "field_type"                                : FIELD_TYPE_B,
            "field_distance_limit"                      : "0.0008",
            "color_metric"                              : "Log Magnitude",
            "alpha_metric"                              : "Magnitude",
            "field_point_scale"                         : "0.0000",
            "field_arrow_head_scale"                    : "0.7500",
            "field_arrow_line_scale"                    : "0.7500",
            "field_boost"                               : "0.0000",
            "display_field_magnitude_labels"            : "True",
            "show_wire_segments"                        : "True",
            "show_wire_points"                          : "True",
            "show_colored_labels"                       : "True",
            "show_gauss"                                : "False",
            "show_coordinate_system"                    : "True",
            "show_perspective_info"                     : "True",
            "dark_background"                           : "True",
            "azimuth"                                   :
                Format.float_to_str(Perspective_Presets.get_by_id("Isometric")["azimuth"]),
            "elevation"                                 :
                Format.float_to_str(Perspective_Presets.get_by_id("Isometric")["elevation"]),
            "scale_factor"                              : "3.0000",
            "constraint_count"                          : "0"
        }



""" Backend_JIT module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Callable, Tuple, Optional
import numpy as np
from numba import jit, prange
from PyQt5.QtCore import QThread
from magneticalc.Backend_Types import get_jit_enabled
from magneticalc.ConditionalDecorator import ConditionalDecorator
from magneticalc.Constants import Constants
from magneticalc.Debug import Debug
from magneticalc.Field_Types import FIELD_TYPE_A, FIELD_TYPE_B


class Backend_JIT:
    """
    Implements the Biot-Savart law for calculating the magnetic flux density (B-field) and vector potential (A-field).
    Backend: JIT.
    """

    def __init__(
            self,
            field_type: int,
            distance_limit: float,
            length_scale: float,
            dc: float,
            current_elements: np.ndarray,
            sampling_volume_points: np.ndarray,
            sampling_volume_permeabilities: np.ndarray,
            progress_callback: Callable
    ) -> None:
        """
        Initializes the class attributes.

        @param field_type: Field type
        @param distance_limit: Distance limit (mitigating divisions by zero)
        @param length_scale: Length scale (m)
        @param dc: Wire current (A)
        @param current_elements: Ordered list of current elements (pairs: [element center, element direction])
        @param sampling_volume_points: Ordered list of sampling volume points
        @param sampling_volume_permeabilities: Ordered list of sampling volume's relative permeabilities µ_r
        @param progress_callback: Progress callback
        """
        Debug(self, ": Init", init=True)

        self._field_type = field_type
        self._distance_limit = distance_limit
        self._length_scale = length_scale
        self._dc = dc
        self._current_elements = current_elements
        self._sampling_volume_points = sampling_volume_points
        self._sampling_volume_permeabilities = sampling_volume_permeabilities
        self._progress_callback = progress_callback

    @staticmethod
    @ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=True)
    def worker(
            field_type: int,
            distance_limit: float,
            length_scale: float,
            current_elements: np.ndarray,
            sampling_volume_point: np.ndarray
    ) -> Optional[Tuple[int, int, np.ndarray]]:
        """
        Applies the Biot-Savart law for calculating the magnetic flux density (B-field) or vector potential (A-field)
        for a single sampling volume point.

        @param field_type: Field type
        @param distance_limit: Distance limit (mitigating divisions by zero)
        @param length_scale: Length scale (m)
        @param current_elements: Ordered list of current elements (pairs: [element center, element direction])
        @param sampling_volume_point: Sampling volume point
        @return: (Total # of calculations, total # of skipped calculations, vector)
        """
        total_calculations = 0
        total_skipped_calculations = 0
        vector = np.zeros(3)

        for j in prange(len(current_elements)):

            element_center = current_elements[j][0]
            element_direction = current_elements[j][1]

            vector_distance = (sampling_volume_point - element_center) * length_scale

            # Calculate distance (mitigating divisions by zero)
            scalar_distance = np.sqrt(vector_distance[0] ** 2 + vector_distance[1] ** 2 + vector_distance[2] ** 2)
            if scalar_distance < distance_limit:
                scalar_distance = distance_limit
                total_skipped_calculations += 1

            total_calculations += 1

            if field_type == FIELD_TYPE_A:
                # Calculate A-field (vector potential)
                vector += element_direction * length_scale / scalar_distance
            elif field_type == FIELD_TYPE_B:
                # Calculate B-field (flux density)
                vector += np.cross(element_direction * length_scale, vector_distance) / (scalar_distance ** 3)

        return total_calculations, total_skipped_calculations, vector

    def get_result(self) -> Optional[Tuple[int, int, np.ndarray]]:
        """
        Calculates the field at every point of the sampling volume.

        @return: (Total # of calculations, total # of skipped calculations, field) if successful, None if interrupted
        """
        Debug(self, ".get_result()")

        total_calculations = 0
        total_skipped_calculations = 0
        vectors = []

        # Fetch resulting vectors
        for i in range(len(self._sampling_volume_points)):

            tup = Backend_JIT.worker(
                self._field_type,
                self._distance_limit,
                self._length_scale,
                self._current_elements,
                self._sampling_volume_points[i]
            )

            total_calculations += tup[0]
            total_skipped_calculations += tup[1]
            vector = tup[2] * self._sampling_volume_permeabilities[i]

            vectors.append(vector)

            # Signal progress update, handle interrupt (every 16 iterations to keep overhead low)
            if i & 0xf == 0:
                self._progress_callback(100 * (i + 1) / len(self._sampling_volume_points))

                if QThread.currentThread().isInterruptionRequested():
                    Debug(self, ".get_result(): WARNING: Interruption requested, exiting now", warning=True)
                    return None

        if self._field_type == FIELD_TYPE_A or self._field_type == FIELD_TYPE_B:
            vectors = np.array(vectors) * self._dc * Constants.mu_0 / 4 / np.pi

        self._progress_callback(100)

        return total_calculations, total_skipped_calculations, vectors



""" Sampling volume module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Tuple, List, Callable
import numpy as np
from PyQt5.QtCore import QThread
from magneticalc.Assert_Dialog import Assert_Dialog
from magneticalc.Constraint import Constraint
from magneticalc.Debug import Debug
from magneticalc.Validatable import Validatable, require_valid, validator


class SamplingVolume(Validatable):
    """ Sampling volume class. """

    def __init__(self) -> None:
        """
        Initializes an empty sampling volume, with zero bounds and no constraints.
        """
        Validatable.__init__(self)
        Debug(self, ": Init", init=True)

        self._bounds_min: np.ndarray = np.zeros(3)
        self._bounds_max: np.ndarray = np.zeros(3)
        self._dimension: Tuple[int, int, int] = (0, 0, 0)

        self._points = np.array([])
        self._permeabilities = np.array([])
        self._labeled_indices: List = []
        self._neighbor_indices: List[np.ndarray] = []

        self.resolution = 0
        self._label_resolution = 0
        self._constraints = []

    def set(
            self,
            resolution: float,
            label_resolution: float,
            constraints: List[Constraint]
    ) -> None:
        """
        Sets the parameters.

        @param resolution: Resolution
        @param label_resolution: Label resolution
        @param constraints: List of constraints
        """
        self.resolution = resolution
        self._label_resolution = label_resolution
        self._constraints = constraints

        Assert_Dialog(resolution > 0, "Resolution must be > 0")
        Assert_Dialog(label_resolution > 0, "Label resolution must be > 0")

    @property
    def bounds(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns this volume's bounding box.

        @return: _bounds_min, _bounds_max
        """
        return self._bounds_min, self._bounds_max

    @property
    def extent(self) -> List:
        """
        Returns this volume's extent.

        @return: 3D point
        """
        return self._bounds_max - self._bounds_min

    def set_bounds_nearest(self, bounds: np.ndarray) -> None:
        """
        Adjusts this volume's bounding box to fully enclose a 3D wire curve.
        This expands the bounding box to the next integer grid coordinates.

        Note: This will not automatically invalidate the sampling volume.

        @param bounds: Bounding box (minimum bounds, maximum bounds)
        @return: Rounded (_bounds_min, _bounds_max)
        """
        self._bounds_min = np.array([np.floor(x) for x in bounds[0]])
        self._bounds_max = np.array([np.ceil(x) for x in bounds[1]])

    def set_padding_nearest(self, padding: np.ndarray) -> None:
        """
        Shrinks or enlarges this volume's bounding box by some amount, in each direction, symmetrically.
        This shrinks or expands the bounding box to the next integer grid coordinates.

        Note: This will not automatically invalidate the sampling volume.

        @param padding: Amount of padding (3D point)
        """
        self._bounds_min -= np.array([np.floor(x) for x in padding])
        self._bounds_max += np.array([np.ceil(x) for x in padding])

    # ------------------------------------------------------------------------------------------------------------------

    def _i_to_xyz(self, _i: int) -> List:
        """
        Convert 1D index to 3D indices.
        Note: "Fortran" indexing (column-major order).

        @param _i: 1D index
        @return: 3D indices
        """
        return [
            _i % self._dimension[0],
            (_i // self._dimension[0]) % self._dimension[1],
            _i // (self._dimension[0] * self._dimension[1])
        ]

    def _xyz_to_i(self, xyz: List) -> int:
        """
        Convert 3D indices to 1D index.
        Note: "Fortran" indexing (column-major order).

        @param xyz: 3D indices
        @return: 1D index
        """
        return xyz[0] + xyz[1] * self._dimension[0] + xyz[2] * self._dimension[0] * self._dimension[1]

    # ------------------------------------------------------------------------------------------------------------------

    @validator
    def recalculate(self, progress_callback: Callable) -> bool:
        """
        Recalculates the sampling volume points, permeabilities, labels and neighborhoods according to the constraints.

        @param progress_callback: Progress callback
        @return: True if successful, False if interrupted
        """
        Debug(self, ".recalculate()")

        # Gather permeabilities by descending permeability
        permeability_groups = sorted(set([constraint.permeability for constraint in self._constraints]), reverse=True)

        # Group constraints by descending permeability
        Permeability_Groups_Precedence_Map = {
            permeability: [constraint for constraint in self._constraints if permeability == constraint.permeability]
            for permeability in permeability_groups
        }
        Debug(
            self,
            ".set(): Permeability groups: { " +
            ", ".join([
                f"µ_r={permeability} ({len(constraints)})"
                for permeability, constraints in Permeability_Groups_Precedence_Map.items()
            ]) + " }"
        )

        # Calculate all possible grid points
        total_axes = [
            np.linspace(
                self._bounds_min[i],
                self._bounds_max[i],
                np.ceil((self._bounds_max[i] - self._bounds_min[i]) * self.resolution).astype(int) + 1
            )
            for i in range(3)
        ]

        self._dimension = np.array([len(axis) for axis in total_axes])
        total_count = np.prod(self._dimension)
        total_points = np.zeros(shape=(total_count, 3))
        total_permeabilities = np.zeros(total_count)
        total_neighbor_indices = np.zeros(shape=(total_count, 6), dtype=int)
        labeled_indices = []

        # Linearly iterate through all possible grid points, computing the 3D cartesian ("euclidean") product
        for i in range(total_count):
            x, y, z = self._i_to_xyz(i)
            point = np.array([total_axes[0][x], total_axes[1][y], total_axes[2][z]])

            point_permeability = 1.0  # Default relative permeability for unconstrained points

            # Iterate over constraints grouped by descending permeability (higher permeabilities take precedence)
            # and calculate the inclusion relation for every group, breaking on the first match.
            for permeability, constraints in Permeability_Groups_Precedence_Map.items():
                if all([constraint.evaluate(point) for constraint in constraints]):
                    point_permeability = permeability
                    break

            if point_permeability != 0:
                # Include this point
                total_points[i] = point
                total_permeabilities[i] = point_permeability

                # Generate this sampling volume point's neighborhood
                neighborhood = [
                    self._xyz_to_i([x + 1, y, z]),
                    self._xyz_to_i([x, y + 1, z]),
                    self._xyz_to_i([x, y, z + 1]),
                    self._xyz_to_i([x - 1, y, z]),
                    self._xyz_to_i([x, y - 1, z]),
                    self._xyz_to_i([x, y, z - 1])
                ]
                total_neighbor_indices[i] = neighborhood

                # Provide orthogonal spacing between labels
                if \
                        np.fmod(x, self.resolution / self._label_resolution) == 0 and \
                        np.fmod(y, self.resolution / self._label_resolution) == 0 and \
                        np.fmod(z, self.resolution / self._label_resolution) == 0:
                    # Generate a label at this point
                    labeled_indices.append([point, i])

            # Signal progress update, handle interrupt (every 256 iterations to keep overhead low)
            if i & 0xff == 0:
                progress_callback(100 * (i + 1) / total_count)

                if QThread.currentThread().isInterruptionRequested():
                    Debug(self, ".recalculate(): WARNING: Interruption requested, exiting now", warning=True)
                    return False

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        index_all_to_filtered = [-1] * total_count
        filtered_index = 0

        # Generate mapping from "all" indices to "filtered" indices
        for i, permeability in enumerate(total_permeabilities):
            if permeability == 0:
                continue

            index_all_to_filtered[i] = filtered_index
            filtered_index += 1

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        points_filtered = []
        permeabilities_filtered = []
        neighbor_indices_filtered = []

        # Filter for included points, i.e. those with permeability != 0; translate neighborhood indices
        for i, permeability in enumerate(total_permeabilities):
            if permeability == 0:
                continue

            point = total_points[i]
            permeability = total_permeabilities[i]

            # Translate neighborhood indices
            neighborhood = total_neighbor_indices[i]
            for j in range(6):
                if 0 <= neighborhood[j] < total_count:
                    if index_all_to_filtered[neighborhood[j]] != -1:
                        neighborhood[j] = index_all_to_filtered[neighborhood[j]]
                    else:
                        neighborhood[j] = -1    # Neighbor out of bounds
                else:
                    neighborhood[j] = -1        # Neighbor out of bounds

            points_filtered.append(point)
            permeabilities_filtered.append(permeability)
            neighbor_indices_filtered.append(neighborhood)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        # Translate label indices
        for i in range(len(labeled_indices)):
            labeled_indices[i][1] = index_all_to_filtered[labeled_indices[i][1]]

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self._points = np.array(points_filtered)
        self._permeabilities = np.array(permeabilities_filtered)
        self._labeled_indices = labeled_indices
        self._neighbor_indices = np.array(neighbor_indices_filtered)

        Debug(
            self,
            ".recalculate(): "
            f"{len(self._constraints)} constraints left {len(self._points)} of {total_count} possible points"
        )

        if len(self._points) == 0:
            Debug(self, ".recalculate: WARNING: Avoiding empty sampling volume by adding origin", warning=True)
            origin = np.zeros(3)
            self._points = np.array([origin])
            self._permeabilities = np.array([0])
            self._labeled_indices = [(origin, 0)]
            self._neighbor_indices = np.array([[0, 0, 0, 0, 0, 0]])

        progress_callback(100)

        return True

    @property
    @require_valid
    def dimension(self) -> Tuple[int, int, int]:
        """
        @return: Sampling volume dimension if it is valid, None otherwise.
        """
        return self._dimension

    @property
    @require_valid
    def points(self) -> np.array:
        """
        Returns this sampling volume's points.

        @return: Array of 3D points
        """
        return self._points

    @property
    @require_valid
    def points_count(self) -> int:
        """
        Returns this sampling volume's point count.

        @return: Point count
        """
        return len(self._points)

    @property
    @require_valid
    def permeabilities(self) -> np.ndarray:
        """
        Returns this sampling volume's relative permeabilities µ_r.

        @return: Ordered list of sampling volume's relative permeabilities µ_r
        """
        return self._permeabilities

    @property
    @require_valid
    def labeled_indices(self) -> List:
        """
        Returns this sampling volume's labeled indices.

        @return: Unordered list of pairs [sampling volume point, field index]
        """
        return self._labeled_indices

    @property
    @require_valid
    def labels_count(self) -> int:
        """
        Returns this sampling volume's label count.

        @return: Label count
        """
        return len(self._labeled_indices)

    @property
    @require_valid
    def neighbor_indices(self) -> List[np.ndarray]:
        """
        Returns this sampling volume's neighborhood indices.

        @return: Ordered list of sampling volume neighborhood indices (six 3D vectors)
        """
        return self._neighbor_indices



""" CalculationThread module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from PyQt5.QtCore import QThread, pyqtSignal
from multiprocessing import cpu_count
from magneticalc.Debug import Debug
from magneticalc.ModelAccess import ModelAccess


class CalculationThread(QThread):
    """
    CalculationThread class.
    The calculation thread implements an offloaded pipeline of subsequent calculation subtasks.
    The pipeline may be interrupted at (almost) any time, thereby exiting this thread.
    """

    # Progress update signal
    _progress_update = pyqtSignal(int)

    # Valid state signals
    _wire_valid             = pyqtSignal()
    _sampling_volume_valid  = pyqtSignal()
    _field_valid            = pyqtSignal()
    _metric_valid           = pyqtSignal()
    _parameters_valid       = pyqtSignal()

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Initializes calculation thread.

        @param gui: GUI
        """
        QThread.__init__(self)
        Debug(self, ": Init", init=True)
        self.gui = gui

        # Connect progress update signal
        self._progress_update.connect(  # type: ignore
            lambda x: self.gui.statusbar.set_progress(x)
        )

        self._wire_valid.connect(self.gui.model.on_wire_valid)  # type: ignore
        self._sampling_volume_valid.connect(self.gui.model.on_sampling_volume_valid)  # type: ignore
        self._field_valid.connect(self.gui.model.on_field_valid)  # type: ignore
        self._metric_valid.connect(self.gui.model.on_metric_valid)  # type: ignore
        self._parameters_valid.connect(self.gui.model.on_parameters_valid)  # type: ignore

    def disconnect_signals(self) -> None:
        """
        Disconnects all valid state signals
        """
        Debug(self, ".disconnect_signals()")
        self._wire_valid.disconnect(self.gui.model.on_wire_valid)  # type: ignore
        self._sampling_volume_valid.disconnect(self.gui.model.on_sampling_volume_valid)  # type: ignore
        self._field_valid.disconnect(self.gui.model.on_field_valid)  # type: ignore
        self._metric_valid.disconnect(self.gui.model.on_metric_valid)  # type: ignore
        self._parameters_valid.disconnect(self.gui.model.on_parameters_valid)  # type: ignore

    def run(self) -> None:
        """
        Thread main function.
        """
        Debug(self, ".run()")

        with ModelAccess(self.gui, recalculate=False):

            if not self.gui.model.wire.valid:
                self.gui.calculation_status.emit("Calculating Wire Segments … (1/5)")

                if not self.gui.model.calculate_wire(
                        self._progress_update.emit  # type: ignore
                ):
                    self._on_finished(False)
                    return

                self._wire_valid.emit()  # type: ignore

            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

            if not self.gui.model.sampling_volume.valid:
                self.gui.calculation_status.emit("Calculating Sampling Volume … (2/5)")

                if not self.gui.model.calculate_sampling_volume(
                        self._progress_update.emit  # type: ignore
                ):
                    self._on_finished(False)
                    return

                self._sampling_volume_valid.emit()  # type: ignore

            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

            if not self.gui.model.field.valid:
                self.gui.calculation_status.emit("Calculating Field … (3/5)")

                num_cores = self.gui.project.get_int("num_cores")
                if num_cores == 0:
                    # "Auto" setting
                    num_cores = max(1, cpu_count() - 1)

                backend_type = self.gui.project.get_int("backend_type")

                success = self.gui.model.calculate_field(
                    progress_callback=self._progress_update.emit,  # type: ignore
                    num_cores=num_cores,
                    backend_type=backend_type
                )

                if not success:
                    self._on_finished(False)
                    return

                self._field_valid.emit()  # type: ignore

            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

            if not self.gui.model.metric.valid:
                self.gui.calculation_status.emit("Calculating Metric … (4/5)")

                if not self.gui.model.calculate_metric(
                        self._progress_update.emit  # type: ignore
                ):
                    self._on_finished(False)
                    return

                self._metric_valid.emit()  # type: ignore

            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

            if not self.gui.model.parameters.valid:
                self.gui.calculation_status.emit("Calculating Parameters … (5/5)")

                if not self.gui.model.calculate_parameters(
                        self._progress_update.emit  # type: ignore
                ):
                    self._on_finished(False)
                    return

                self._parameters_valid.emit()  # type: ignore

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self._on_finished(True)

    # ------------------------------------------------------------------------------------------------------------------

    def _on_finished(self, success: bool) -> None:
        """
        Signals that the calculation finished.

        @param success: True if calculation was successful, False otherwise
        """
        Debug(self, f".on_finished(success={success})")

        # Firing this signal results in slightly delayed execution, even after joining this thread;
        # the execution of "on_calculation_exited()" will later be skipped when it sees another thread running.
        self.gui.calculation_exited.emit(success)



""" Usage_Dialog module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from magneticalc.QtWidgets2.QDialog2 import QDialog2
from magneticalc.QtWidgets2.QTextBrowser2 import QTextBrowser2
from magneticalc.Debug import Debug
from magneticalc.Theme import Theme


class Usage_Dialog(QDialog2):
    """ Usage_Dialog class. """

    # HTML content
    HTML = f"""
        <h3 style="color: {Theme.MainColor};">First Steps</h3>
        <h4>
            MagnetiCalc loads a very simple magnetic setup by default.<br>
            Follow these steps to turn it into your own custom setup:
        </h4>
        <ul>
            <li>
                Go to <i>Wire</i> › <i>Load Preset</i> to select a basic wire shape.<br>
                (But you could also import/export wire points from/to a TXT file.)
            </li>
        </ul>
        <ul>
            <li>
                Customize the wire's base points inside the spreadsheet.<br>
                Apply some transformations to your basic wire; stretch its shape, rotate or duplicate it.<br>
                Reduce the slicer limit to improve the field calculation accuracy.<br>
                Set the electrical current flowing through your wire.
            </li>
        </ul>
        <ul>
            <li>
                The sampling volume is the grid over which the field is calculated;<br>
                increase its resolution to the desired level to show every detail of the field.<br>
                <i>Warning:</i> Detailed sampling volumes can take a considerable amount of time to compute.<br>
                <b>Please note that you can disable Auto-Calculation at any time.<br>
                You can and hit F5 or ESC to start or stop the calculation on demand.</b>
            </li>
        </ul>
        <ul>
            <li>
                By default, the cuboid sampling volume covers the wire completely;<br>
                however, you can symmetrically adjust its bounding box by adding (subtracting) some padding.<br>
                By setting a <i>negative</i> padding, you can reduce the sampling volume to a plane, line or point.<br>
                <i>Experimental Feature:</i>
                Use the constraint editor to create regions of relative permeability µ<sub>r</sub> ≠ 1.<br>
                Constraints can also be used to selectively <i>disable</i> calculation over some regions.
            </li>
        </ul>
        <ul>
            <li>
                Select the type of field to calculate: A-field (vector potential) or B-field (flux density).<br>
                <i>Note:</i> Current element center points may be located very close to sampling volume points;<br>
                as this distance approaches zero, the field magnitude approaches infinity (singular behaviour).<br>
                Therefore, some distance limit must be set to cut off (skip) these infinities during calculation.
            </li>
        </ul>
        <ul>
            <li>
                Select a color/alpha metric to adjust the field's hue/transparency individually.<br>
                Because all metrics are normalized, changes in DC current do not affect the field color/alpha;<br>
                however, the displayed metric limits linearly scale with the DC current.
            </li>
        </ul>
        <ul>
            <li>
                Take a look at the minimum and maximum magnetic flux densities reached in each metric.<br>
                <i>Note:</i> To achieve an accurate coil energy and self-inductance calculation,<br>
                ensure that the sampling volume encloses a large, non-singular portion of the field.
            </li>
        </ul>
        <ul>
            <li>
                Use the scroll wheel to zoom in and out of the 3D scene.<br>
                Click and drag into the 3D scene to rotate.<br>
                Press SHIFT while dragging to <i>move</i> the entire 3D scene.<br>
                If you like the result, press CTRL+I to save a screenshot.<br>
                You can also export the fields, wire and current to an HDF5 container for use in post-processing.<br>
                Finally, make sure to save your project!
            </li>
        </ul>

        <h3 style="color: {Theme.MainColor};">What does MagnetiCalc do?</h3>

        MagnetiCalc calculates the static magnetic flux density, vector potential, energy, self-inductance
        and magnetic dipole moment of arbitrary coils.
        Inside an OpenGL-accelerated GUI, the magnetic flux density
        (B-field, in units of <i>Tesla</i>)
        or the magnetic vector potential (A-field, in units of <i>Tesla-meter</i>)
        is displayed in interactive 3D, using multiple metrics for highlighting the field properties.<br><br>

        <i>Experimental feature:</i> To calculate the energy and self-inductance of permeable (i.e. ferrous)
        materials, different core media can be modeled as regions of variable relative permeability;
        however, core saturation is currently not modeled, resulting in excessive flux density values.

        <h3 style="color: {Theme.MainColor};">Who needs MagnetiCalc?</h3>

        MagnetiCalc does its job for hobbyists, students, engineers and researchers of magnetic phenomena.
        I designed MagnetiCalc from scratch, because I didn't want to mess around with expensive and/or
        overly complex simulation software whenever I needed to solve a magnetostatic problem.

        <h3 style="color: {Theme.MainColor};">How does it work?</h3>

        The B-field calculation is implemented using the Biot-Savart law [<b>1</b>],
        employing multiprocessing techniques;
        MagnetiCalc uses just-in-time compilation (JIT) and, if available, GPU-acceleration (CUDA)
        to achieve high-performance calculations.
        Additionally, the use of easily constrainable "sampling volumes"
        allows for selective calculation over grids of arbitrary shape and arbitrary relative permeabilities
        <i>(experimental)</i>.<br><br>

        The shape of any wire is modeled as a 3D piecewise linear curve.
        Arbitrary loops of wire are sliced into differential current elements, each of which
        contributes to the total field magnitude at some fixed 3D grid point,
        summing over the positions of all current elements.<br><br>

        At each grid point, the field is displayed using colored arrows and/or dots;
        field color and alpha transparency are individually mapped using one of the various available metrics.<br><br>

        The coil's energy [<b>2</b>] and self-inductance [<b>3</b>]
        are calculated by summing the squared B-field over the entire sampling volume;
        ensure that the sampling volume encloses a large, non-singular portion of the field.<br><br>

        Additionally, the scalar magnetic dipole moment [<b>4</b>]
        is calculated by summing over all current elements.<br><br>

        [<b>1</b>]: Jackson, Klassische Elektrodynamik, 5. Auflage, S. 204, (5.4).<br>
        [<b>2</b>]: Kraus, Electromagnetics, 4th Edition, p. 269, 6-9-1.<br>
        [<b>3</b>]: Jackson, Klassische Elektrodynamik, 5. Auflage, S. 252, (5.157).<br>
        [<b>4</b>]: Jackson, Klassische Elektrodynamik, 5. Auflage, S. 216, (5.54).<br>

        <br><br><span style="color: {Theme.MainColor};">
            This and more information about MagnetiCalc can be found in the <code>README.md</code> file and on GitHub.
        </span><br>
        """

    def __init__(self) -> None:
        """
        Initializes "Usage" dialog.
        """
        QDialog2.__init__(self, title="Usage", width=850)
        Debug(self, ": Init", init=True)

        text_browser = QTextBrowser2(html=self.HTML)
        text_browser.setMinimumHeight(600)
        self.addWidget(text_browser)
        self.addButtons({
            "OK": ("fa.check", self.accept)
        })



""" ExportContainer_Dialog module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from PyQt5.QtWidgets import QCheckBox
from magneticalc.QtWidgets2.QLabel2 import QLabel2
from magneticalc.QtWidgets2.QDialog2 import QDialog2
from magneticalc.QtWidgets2.QSaveAction import QSaveAction
from magneticalc.API import API
from magneticalc.Debug import Debug
from magneticalc.Field_Types import Field_Types_Names_Map, Field_Types_Abbreviations_Map, field_name_to_type
from magneticalc.Theme import Theme


class ExportContainer_Dialog(QDialog2):
    """ ExportContainer_Dialog class. """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Initializes the dialog.

        @param gui: GUI
        """
        QDialog2.__init__(self, title="Export Container", width=500)
        Debug(self, ": Init", init=True)
        self.gui = gui

        self.addWidget(QLabel2("Please select items for export", bold=True, color=Theme.MainColor))
        self.addSpacing(8)
        self.addWidget(QLabel2(
            "Fields must have been calculated before they can be exported.", italic=True, color=Theme.LiteColor
        ))
        self.addSpacing(16)

        self.Export_Names_Available_Map = {
            "Wire Points"   : self.gui.model.wire.valid,
            "Wire Current"  : self.gui.model.wire.valid
        }
        self.Export_Names_Available_Map.update({
            field_name: self.gui.model.get_valid_field(field_type) is not None
            for field_type, field_name in Field_Types_Names_Map.items()
        })

        self.checkboxes = {}
        for i, item in enumerate(self.Export_Names_Available_Map.items()):
            item_name, item_available = item
            self.checkboxes[item_name] = QCheckBox(" " + item_name + ("" if item_available else " (not calculated)"))
            self.checkboxes[item_name].setEnabled(item_available)
            self.checkboxes[item_name].setChecked(item_available)
            self.addWidget(self.checkboxes[item_name])

        self.addSpacing(16)

        buttons = self.addButtons({
            "Cancel"            : ("fa.close", self.reject),
            "Save Container …"  : ("fa.save", self.export)
        })
        buttons[1].setFocus()

    def export(self) -> None:
        """
        Exports wire points, current and fields to some HDF5 container file.
        """
        Debug(self, ".export()")

        Export_Names_Selection_Map = {
            item_name: self.checkboxes[item_name].isChecked()
            for item_name, item_available in self.Export_Names_Available_Map.items()
        }

        Export_Names_Abbreviations_Map = {
            "Wire Current"  :   "Current",
            "Wire Points"   :   "Wire"
        }
        Export_Names_Abbreviations_Map.update({
            field_name: Field_Types_Abbreviations_Map[field_type]
            for field_type, field_name in Field_Types_Names_Map.items()
        })
        export_abbreviations_str = "-".join([
            Export_Names_Abbreviations_Map[item_name]
            for item_name, item_selected in Export_Names_Selection_Map.items() if item_selected
        ])

        action = QSaveAction(
            self.gui,
            title="Export Container",
            date=True,
            filename="MagnetiCalc_Export" + (("_" + export_abbreviations_str) if export_abbreviations_str else ""),
            extension=".hdf5",
            _filter="HDF5 Container (*.hdf5)"
        )
        if not action.filename:
            self.reject()
            return

        container = {}
        fields = {}

        for item_name, item_selected in Export_Names_Selection_Map.items():

            if not item_selected:
                continue

            if item_name == "Wire Points":
                wire_points_components = self.gui.model.wire.points_sliced.T
                wire_points = dict(zip(["x", "y", "z"], wire_points_components))
                container.update({"wire_points": wire_points})

            elif item_name == "Wire Current":
                wire_current = self.gui.model.wire.dc
                container.update({"wire_current": wire_current})

            else:
                field_type = field_name_to_type(item_name)
                field_components = self.gui.model.get_valid_field(field_type).vectors.T
                field_abbreviation = Field_Types_Abbreviations_Map[field_type]
                fields.update(dict(zip(
                    [field_abbreviation + "_x", field_abbreviation + "_y", field_abbreviation + "_z"],
                    field_components
                )))

        if fields != {}:
            fields.update(dict(zip(["nx", "ny", "nz"], self.gui.model.sampling_volume.dimension)))
            fields.update(dict(zip(["x", "y", "z"], self.gui.model.sampling_volume.points.T)))
            container.update({"fields": fields})

        API.export_hdf5(action.filename, container)

        self.accept()



""" API module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import h5py
import numpy as np
from typing import Dict, List, Union
from magneticalc.MagnetiCalc_Data import MagnetiCalc_Data


class API:
    """ API class. """

    @staticmethod
    def import_wire(filename: str) -> np.ndarray:
        """
        Imports wire points from a TXT file.

        @param filename: Filename
        @return: NumPy array of 3D points
        """
        data = np.loadtxt(filename)
        assert data.shape[1] == 3, "Expecting array of 3D points"
        return data

    @staticmethod
    def export_wire(filename: str, data: Union[List, np.ndarray]) -> None:
        """
        Exports wire points to a TXT file.

        @param filename: Filename
        @param data: NumPy array of 3D points
        """
        _data_ = np.array(data)
        assert _data_.shape[1] == 3, "Expecting array of 3D points"
        np.savetxt(
            filename,
            _data_  # type: ignore
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    @staticmethod
    def import_hdf5(filename: str) -> MagnetiCalc_Data:
        """
        Imports data from an HDF5 container.

        Opens an HDF5 file and converts every group and subgroup into a dictionary
        where the keys are the group keys and the items are the datasets.

        @param filename: Filename
        @return: MagnetiCalc_Data object (can be accessed like a dictionary)
        """
        hdf5_group = h5py.File(filename, "r")
        data = {}
        API._hdf5_group_to_dict(hdf5_group, data)
        hdf5_group.close()
        return MagnetiCalc_Data(data)

    @staticmethod
    def export_hdf5(filename: str, data: Union[Dict, MagnetiCalc_Data]) -> None:
        """
        Exports data to an HDF5 container.

        Takes a dictionary and writes an HDF5 file using keys as keys,
        and items as groups if they are dictionaries or as datasets otherwise.

        @param filename: Filename
        @param data: Dictionary or MagnetiCalc_Data object
        """
        hdf5_group = h5py.File(filename, "w")
        _data_ = data.dictionary if isinstance(data, MagnetiCalc_Data) else data
        API._dict_to_hdf5_group(hdf5_group, _data_)
        hdf5_group.close()

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def _dict_to_hdf5_group(hdf5_group: h5py.Group, dictionary: Dict) -> None:
        """
        Recursively transforms a dictionary into an HDF5 group (in-place).

        @param hdf5_group: HDF5 group
        @param dictionary: Dictionary
        """
        for key in dictionary.keys():
            if isinstance(dictionary[key], dict):
                group = hdf5_group.create_group(key)
                API._dict_to_hdf5_group(group, dictionary[key])
            else:
                hdf5_group[key] = dictionary[key]

    @staticmethod
    def _hdf5_group_to_dict(hdf5_group: h5py.Group, dictionary: Dict) -> None:
        """
        Recursively transforms an HDF5 group into a dictionary (in-place).

        @param hdf5_group: HDF5 group
        @param dictionary: Dictionary
        """
        for key in [key for key in hdf5_group]:
            if isinstance(hdf5_group[key], h5py.Dataset):
                dictionary[key] = hdf5_group[key][()]
            else:
                dictionary[key] = {}
                API._hdf5_group_to_dict(hdf5_group[key], dictionary[key])



""" QGroupBox2 module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from PyQt5.QtWidgets import QGroupBox
from magneticalc.QtWidgets2.QLayouted import QLayouted


class QGroupBox2(QGroupBox, QLayouted):
    """ QGroupBox2 class. """

    def __init__(self, title: str, color: str) -> None:
        """
        Initializes a groupbox.

        @param title: Title
        """
        QGroupBox.__init__(self)
        QLayouted.__init__(self)
        self.install_layout(self)

        self.setTitle(title)

        self.set_color(color)

    def set_color(self, color: str) -> None:
        """
        Sets the style and label color of the groupbox.

        @param color: Label color
        """
        self.setStyleSheet(f"""
            QGroupBox {{
                border: 1px solid #cccccc;
                border-radius: 3px;
                margin-top: 20px;
                color: {color};
                font-weight: bold;
                background-color: #e5e5e5;
            }}

            QGroupBox::title {{
                subcontrol-origin: margin;
                subcontrol-position: top left;
                background-color: palette(window);
            }}
        """)



""" Help_Menu module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from functools import partial
import webbrowser
import qtawesome as qta
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QMenu
from magneticalc.About_Dialog import About_Dialog
from magneticalc.CheckForUpdates_Dialog import CheckForUpdates_Dialog
from magneticalc.Debug import Debug
from magneticalc.Usage_Dialog import Usage_Dialog
from magneticalc.Version import __URL__


class Help_Menu(QMenu):
    """ Help_Menu class. """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ):
        """
        Initializes the menu.
        """
        QMenu.__init__(self, "&Help")
        Debug(self, ": Init", init=True)
        self.gui = gui

        self.addAction(qta.icon("fa.info"), "&Usage …", lambda: Usage_Dialog().show(), Qt.Key_F1)
        self.addSeparator()
        self.addAction(
            qta.icon("fa.newspaper-o"), "&Check for Updates …", lambda: CheckForUpdates_Dialog().show()
        )
        self.addAction(
            qta.icon("fa.github"),
            "&GitHub Repository …",
            partial(webbrowser.open, __URL__)  # type: ignore
        )
        self.addSeparator()
        self.addAction(qta.icon("fa.coffee"), "&About …", lambda: About_Dialog().show())

    def update(self):
        """
        Updates the menu.
        """
        Debug(self, ".update()", refresh=True)
